# -*- coding: utf-8 -*-
from qtpy import QtCore, QtWidgets
from ..custom import FloatLineEdit


class DiamondWidget(QtWidgets.QWidget):
    def __init__(self, *args):
        super(DiamondWidget, self).__init__(*args)

        self.create_widgets()
        self.style_widgets()
        self.create_layout()

    def create_widgets(self):
        self.diamond_optimize_btn = QtWidgets.QPushButton("Optimize")
        self.diamond_lbl = QtWidgets.QLabel('Amount:')
        self.diamond_txt = FloatLineEdit('0')

    def style_widgets(self):
        self.diamond_optimize_btn.setFlat(True)
        self.diamond_lbl.setAlignment(
            QtCore.Qt.AlignVCenter | QtCore.Qt.AlignRight)

        self.diamond_txt.setMaximumWidth(60)

    def create_layout(self):
        self._layout = QtWidgets.QHBoxLayout()
        self._layout.addWidget(self.diamond_lbl)
        self._layout.addWidget(self.diamond_txt)
        self._layout.addWidget(self.diamond_optimize_btn)

        self.setLayout(self._layout)



# -*- coding: utf-8 -*-

from qtpy import QtWidgets, QtCore

from ..custom import FlatButton, HorizontalLine, LabelAlignRight


class TransferFunctionWidget(QtWidgets.QWidget):
    def __init__(self, *args):
        super(TransferFunctionWidget, self).__init__(*args)

        self.create_widgets()
        self.create_layout()
        self.style_widgets()
        self.create_signals()

    def create_widgets(self):
        self.load_std_btn = FlatButton("Load Std")
        self.load_std_bkg_btn = FlatButton("Load Std Bkg")
        self.load_sample_btn = FlatButton("Load Sample")
        self.load_sample_bkg_btn = FlatButton("Load Sample Bkg")

        self.std_filename_lbl = LabelAlignRight('')
        self.std_bkg_filename_lbl = LabelAlignRight("")
        self.sample_filename_lbl = LabelAlignRight("")
        self.sample_bkg_filename_lbl = LabelAlignRight("")

        self.std_bkg_scaling_sb = QtWidgets.QDoubleSpinBox()
        self.std_bkg_scaling_sb.setValue(1.0)
        self.std_bkg_scaling_sb.setSingleStep(0.01)

        self.sample_bkg_scaling_sb = QtWidgets.QDoubleSpinBox()
        self.sample_bkg_scaling_sb.setValue(1.0)
        self.sample_bkg_scaling_sb.setSingleStep(0.01)

        self.smooth_sb = QtWidgets.QDoubleSpinBox()
        self.smooth_sb.setValue(1.0)
        self.smooth_sb.setSingleStep(0.1)

    def create_layout(self):
        self.main_layout = QtWidgets.QVBoxLayout()

        self.activate_cb = QtWidgets.QCheckBox("activate")
        self.main_layout.addWidget(self.activate_cb)
        self.main_layout.addWidget(HorizontalLine())

        self.transfer_layout = QtWidgets.QGridLayout()
        self.transfer_layout.addWidget(self.load_sample_btn, 0, 0)
        self.transfer_layout.addWidget(self.sample_filename_lbl, 0, 1)
        self.transfer_layout.addWidget(self.load_sample_bkg_btn, 1, 0)
        self.transfer_layout.addWidget(self.sample_bkg_filename_lbl, 1, 1)

        self.transfer_layout.addWidget(self.load_std_btn, 2, 0)
        self.transfer_layout.addWidget(self.std_filename_lbl, 2, 1)
        self.transfer_layout.addWidget(self.load_std_bkg_btn, 3, 0)
        self.transfer_layout.addWidget(self.std_bkg_filename_lbl, 3, 1)

        self.scaling_gb = QtWidgets.QGroupBox("")
        self.scaling_layout = QtWidgets.QGridLayout()
        self.scaling_layout.addItem(QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.MinimumExpanding,
                                                          QtWidgets.QSizePolicy.Fixed), 0, 0)
        self.scaling_layout.addWidget(LabelAlignRight("Sample bkg scaling:"), 0, 1)
        self.scaling_layout.addWidget(self.sample_bkg_scaling_sb, 0, 2)
        self.scaling_layout.addWidget(LabelAlignRight("Std bkg scaling:"), 1, 1)
        self.scaling_layout.addWidget(self.std_bkg_scaling_sb, 1, 2)
        self.scaling_layout.addWidget(LabelAlignRight("Smoothing:"), 2, 1)
        self.scaling_layout.addWidget(self.smooth_sb, 2, 2)

        self.scaling_gb.setLayout(self.scaling_layout)
        self.transfer_layout.addWidget(self.scaling_gb, 4, 0, 1, 2)

        self.main_layout.addLayout(self.transfer_layout)
        self.setLayout(self.main_layout)

    def style_widgets(self):
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(5)

        self.transfer_layout.setContentsMargins(5, 5, 5, 5)

        self.sample_bkg_scaling_sb.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.std_bkg_scaling_sb.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.smooth_sb.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)

        self.sample_bkg_scaling_sb.setMinimumWidth(75)
        self.std_bkg_scaling_sb.setMinimumWidth(75)
        self.smooth_sb.setMinimumWidth(75)

    def create_signals(self):
        pass



# -*- coding: utf-8 -*-

from qtpy import QtWidgets


class ExpandableBox(QtWidgets.QWidget):
    def __init__(self, content_widget, title='', hide=False):
        super(ExpandableBox, self).__init__()

        self._vertical_layout = QtWidgets.QVBoxLayout()
        self._vertical_layout.setContentsMargins(0, 0, 0, 0)
        self._vertical_layout.setSpacing(0)

        self.create_head_widget(title)
        self.create_content_widget(content_widget)
        self.set_custom_stylesheet()

        self._vertical_layout.addWidget(self.head_widget)
        self._vertical_layout.addWidget(self.content_widget)
        self.setLayout(self._vertical_layout)

        self.minimize_btn.clicked.connect(self.change_visible_state)

        self.minimized = False

        if hide:
            self.change_visible_state()

    def create_head_widget(self, title):
        self._head_layout = QtWidgets.QHBoxLayout()
        self._head_layout.setContentsMargins(10, 8, 15, 3)
        self._head_layout.setSpacing(0)

        self.minimize_btn = QtWidgets.QPushButton("-")
        self.minimize_btn.setFixedHeight(20)
        self.minimize_btn.setFixedWidth(20)
        self.minimize_btn.setObjectName("minimize_btn")

        self.title_lbl = QtWidgets.QLabel(title)
        self.title_lbl.setStyleSheet("font: italic 15px;")

        self._head_layout.addWidget(self.minimize_btn)
        self._head_layout.addSpacing(10)
        self._head_layout.addWidget(self.title_lbl)

        self.head_widget = QtWidgets.QWidget()
        self.head_widget.setLayout(self._head_layout)
        self.head_widget.setObjectName("head_widget")

    def create_content_widget(self, content_widget):

        self._content_widget = content_widget

        self.content_widget = QtWidgets.QWidget()
        self.content_layout = QtWidgets.QVBoxLayout()
        self.content_layout.setContentsMargins(8, 8, 8, 8)
        self.content_layout.setSpacing(0)
        self.content_layout.addWidget(self._content_widget)
        self.content_widget.setLayout(self.content_layout)

        self.content_widget.setObjectName("content_widget")

    def set_custom_stylesheet(self):
        self.setStyleSheet(
            """
            QLabel, #head_widget{
                background: #303030;
            }
            #head_widget{
                border-top-left-radius: 10px;
                border-top-right-radius: 10px;
                border: 2px solid #666;
                border-bottom: 1px solid #666;
            }
            """
        )
        self.content_widget.setStyleSheet(
            """
            QLabel, QGroupBox, #content_widget{
                background: #3B3B3B;
            }
            #content_widget{
                border-bottom-left-radius: 10px;
                border-bottom-right-radius: 10px;
                border: 2px solid #666;
                border-top: None;
            }
            """
        )
        self.head_widget.setStyleSheet(
            """
            #minimize_btn{
                border-radius:0px;
                padding: 0px;
                margin: 0px;
                margin-bottom: 2px;
                margin-left: 2px;
                border-top-left-radius: 5px;
            }
            """
        )

    def change_visible_state(self):
        if self.minimized:
            self.content_widget.show()
            self.minimized = False
            self.minimize_btn.setText("-")
        else:
            self.content_widget.hide()
            self.minimized = True
            self.minimize_btn.setText("+")



# -*- coding: utf-8 -*-

from datetime import date, datetime, timedelta
from .configuration import GlassureConfiguration, ExtrapolationConfiguration, TransformConfiguration, Sample
import numpy as np
import json
from lmfit import Parameters, minimize
from qtpy import QtGui, QtCore

from ...core.pattern import Pattern
from ...core.utility import calculate_incoherent_scattering, convert_density_to_atoms_per_cubic_angstrom
from ...core import calculate_sq, calculate_gr, calculate_fr
from ...core.optimization import optimize_sq
from ...core.soller_correction import SollerCorrectionGui
from ...core.transfer_function import calculate_transfer_function

from ...core.utility import extrapolate_to_zero_linear, extrapolate_to_zero_step, extrapolate_to_zero_spline, \
    extrapolate_to_zero_poly, calculate_s0

from ...core.scattering_factors import get_available_elements


class GlassureModel(QtCore.QObject):
    configurations_changed = QtCore.Signal()
    configuration_selected = QtCore.Signal(int)
    data_changed = QtCore.Signal()
    sq_changed = QtCore.Signal(Pattern)
    fr_changed = QtCore.Signal(Pattern)
    gr_changed = QtCore.Signal(Pattern)

    def __init__(self):
        super(GlassureModel, self).__init__()

        self.configurations = []
        self.configurations.append(GlassureConfiguration())
        self.configuration_ind = 0

        self.auto_update = True
        self.optimization_callback = None

    def load_data(self, filename):
        self.original_pattern.load(filename)
        self.calculate_transforms()

    def load_bkg(self, filename):
        self.current_configuration.background_pattern = Pattern.from_file(filename)
        self.calculate_transforms()

    def reset_bkg(self):
        self.current_configuration.background_pattern = None
        self.calculate_transforms()

    @property
    def current_configuration(self):
        """
        :rtype: GlassureConfiguration
        """
        return self.configurations[self.configuration_ind]

    def add_configuration(self):
        self.configurations.append(self.current_configuration.copy())
        self.configuration_ind = len(self.configurations) - 1
        self.configurations_changed.emit()

    def remove_configuration(self):
        """
        Removes the currently selected configuration, unless only one
        configuration is left
        """
        if len(self.configurations) == 1:
            return

        del self.configurations[self.configuration_ind]

        if self.configuration_ind >= len(self.configurations):
            self.configuration_ind = len(self.configurations) - 1
        self.configurations_changed.emit()

    def select_configuration(self, ind):
        if ind < 0:
            ind = len(self.configurations) + ind
        self.configuration_ind = ind
        self.configuration_selected.emit(ind)

    def to_json(self, filename):
        data = [c.to_dict() for c in self.configurations]
        with open(filename, 'w') as f:
            json.dump(data, f)

    def read_json(self, filename):
        with open(filename, 'r') as f:
            data = json.load(f)
        self.configurations = [GlassureConfiguration.from_dict(d)
                               for d in data]
        self.configuration_ind = 0
        self.configurations_changed.emit()

    @property
    def atomic_density(self):
        if len(self.composition):
            return convert_density_to_atoms_per_cubic_angstrom(self.composition, self.density)
        return 0

    @property
    def original_pattern(self):
        return self.current_configuration.original_pattern

    @original_pattern.setter
    def original_pattern(self, new_pattern):
        self.current_configuration.original_pattern = new_pattern

    @property
    def background_pattern(self):
        if self.current_configuration.diamond_bkg_pattern is None:
            if self.current_configuration.background_pattern is None:
                return None
            else:
                return self.current_configuration.background_pattern
        if self.current_configuration.background_pattern is None:
            return self.current_configuration.diamond_bkg_pattern
        return self.current_configuration.background_pattern + \
            self.current_configuration.diamond_bkg_pattern

    @property
    def diamond_bkg_pattern(self):
        return self.current_configuration.diamond_bkg_pattern

    @diamond_bkg_pattern.setter
    def diamond_bkg_pattern(self, new_pattern):
        self.current_configuration.diamond_bkg_pattern = new_pattern

    @property
    def background_scaling(self):
        return self.current_configuration.background_pattern.scaling

    @background_scaling.setter
    def background_scaling(self, new_value):
        self.current_configuration.background_pattern.scaling = new_value
        self.calculate_transforms()

    @property
    def sq_pattern(self):
        return self.current_configuration.sq_pattern

    @sq_pattern.setter
    def sq_pattern(self, new_sq):
        self.current_configuration.sq_pattern = new_sq
        self.sq_changed.emit(new_sq)

    @property
    def fr_pattern(self):
        return self.current_configuration.fr_pattern

    @fr_pattern.setter
    def fr_pattern(self, new_fr):
        self.current_configuration.fr_pattern = new_fr
        self.fr_changed.emit(new_fr)

    @property
    def gr_pattern(self):
        return self.current_configuration.gr_pattern

    @gr_pattern.setter
    def gr_pattern(self, new_gr):
        self.current_configuration.gr_pattern = new_gr
        self.gr_changed.emit(new_gr)

    @property
    def sample(self):
        return self.current_configuration.sample

    @sample.setter
    def sample(self, new_sample):
        self.current_configuration.sample = new_sample
        self.calculate_transforms()

    def add_element(self):
        elements = get_available_elements(self.sample.sf_source)
        for element in elements:
            if element not in self.sample.composition.keys():
                self.sample.composition[element] = 1
                break
        self.calculate_transforms()

    @property
    def sf_source(self):
        return self.sample.sf_source

    @sf_source.setter
    def sf_source(self, new_source):
        self.sample.sf_source = new_source
        self.calculate_transforms()

    @property
    def composition(self):
        return self.sample.composition

    @composition.setter
    def composition(self, new_composition):
        self.sample.composition = new_composition
        self.calculate_transforms()

    @property
    def density(self):
        return self.sample.density

    @density.setter
    def density(self, new_density):
        self.sample.density = new_density
        self.calculate_transforms()

    @property
    def density_error(self):
        return self.sample.density_error

    @density_error.setter
    def density_error(self, new_density_error):
        self.sample.density_error = new_density_error

    @property
    def transform_config(self) -> TransformConfiguration:
        return self.current_configuration.transform_config

    @transform_config.setter
    def transform_config(self, new_transform_config: TransformConfiguration):
        self.current_configuration.transform_config = new_transform_config
        self.calculate_transforms()

    @property
    def q_min(self):
        return self.transform_config.q_min

    @q_min.setter
    def q_min(self, new_q_min):
        self.transform_config.q_min = new_q_min
        self.calculate_transforms()

    @property
    def q_max(self):
        return self.transform_config.q_max

    @q_max.setter
    def q_max(self, new_q_max):
        self.transform_config.q_max = new_q_max
        self.calculate_transforms()

    @property
    def r_min(self):
        return self.transform_config.r_min

    @r_min.setter
    def r_min(self, new_r_min):
        self.transform_config.r_min = new_r_min
        self.calculate_transforms()

    @property
    def r_max(self):
        return self.transform_config.r_max

    @r_max.setter
    def r_max(self, new_r_max):
        self.transform_config.r_max = new_r_max
        self.calculate_transforms()

    @property
    def r_step(self):
        return self.transform_config.r_step

    @r_step.setter
    def r_step(self, new_r_step):
        self.transform_config.r_step = new_r_step
        self.calculate_transforms()

    @property
    def optimize(self):
        return self.current_configuration.optimize_config.enable

    @optimize.setter
    def optimize(self, new_flag):
        self.current_configuration.optimize_config.enable = new_flag
        self.calculate_transforms()

    @property
    def r_cutoff(self):
        return self.current_configuration.optimize_config.r_cutoff

    @r_cutoff.setter
    def r_cutoff(self, new_r_cutoff):
        self.current_configuration.optimize_config.r_cutoff = new_r_cutoff
        self.calculate_transforms()

    @property
    def optimization_iterations(self):
        return self.current_configuration.optimize_config.iterations

    @optimization_iterations.setter
    def optimization_iterations(self, new_value):
        self.current_configuration.optimize_config.iterations = new_value

    @property
    def optimization_attenuation(self):
        return self.current_configuration.optimize_config.attenuation

    @optimization_attenuation.setter
    def optimization_attenuation(self, new_value):
        self.current_configuration.optimize_config.attenuation = new_value

    @property
    def use_modification_fcn(self):
        return self.transform_config.use_modification_fcn

    @use_modification_fcn.setter
    def use_modification_fcn(self, value):
        self.transform_config.use_modification_fcn = value
        self.calculate_transforms()

    @property
    def sq_method(self):
        return self.transform_config.sq_method

    @sq_method.setter
    def sq_method(self, value):
        self.transform_config.sq_method = value
        self.calculate_transforms()

    @property
    def normalization_method(self):
        return self.transform_config.normalization_method

    @normalization_method.setter
    def normalization_method(self, value):
        self.transform_config.normalization_method = value
        self.calculate_transforms()

    @property
    def extrapolation_config(self) -> ExtrapolationConfiguration:
        return self.current_configuration.extrapolation_config

    @extrapolation_config.setter
    def extrapolation_config(self, value: ExtrapolationConfiguration):
        self.current_configuration.extrapolation_config = value
        self.calculate_transforms()

    @property
    def use_soller_correction(self):
        return self.current_configuration.soller_config.enable

    @use_soller_correction.setter
    def use_soller_correction(self, value):
        self.current_configuration.soller_config.enable = value
        self.calculate_transforms()

    @property
    def soller_correction(self):
        return self.current_configuration.soller_config.correction

    @soller_correction.setter
    def soller_correction(self, new_value):
        self.current_configuration.soller_config.correction = new_value
        self.calculate_transforms()

    @property
    def soller_parameters(self):
        return self.current_configuration.soller_config.parameters

    @soller_parameters.setter
    def soller_parameters(self, new_parameters):
        self.current_configuration.soller_config.parameters = new_parameters
        self.calculate_transforms()

    @property
    def use_transfer_function(self):
        return self.current_configuration.transfer_config.enable

    @use_transfer_function.setter
    def use_transfer_function(self, new_value):
        self.current_configuration.transfer_config.enable = new_value
        if new_value:
            self.update_transfer_function()
        else:
            self.calculate_transforms()

    @property
    def transfer_function(self):
        return self.current_configuration.transfer_config.function

    @property
    def transfer_function_smoothing(self):
        return self.current_configuration.transfer_config.smoothing

    @transfer_function_smoothing.setter
    def transfer_function_smoothing(self, new_value):
        self.current_configuration.transfer_config.smoothing = new_value
        self.update_transfer_function()

    @property
    def transfer_std_pattern(self):
        """
        :rtype: Pattern
        """
        return self.current_configuration.transfer_config.std_pattern

    @transfer_std_pattern.setter
    def transfer_std_pattern(self, new_pattern):
        self.current_configuration.transfer_config.std_pattern = new_pattern

    @property
    def transfer_std_bkg_pattern(self):
        """
        :rtype Pattern:
        """
        return self.current_configuration.transfer_config.std_bkg_pattern

    @transfer_std_bkg_pattern.setter
    def transfer_std_bkg_pattern(self, new_pattern):
        self.current_configuration.transfer_config.std_bkg_pattern = new_pattern
        self.update_transfer_function()

    @property
    def transfer_std_bkg_scaling(self):
        return self.current_configuration.transfer_config.std_bkg_scaling

    @transfer_std_bkg_scaling.setter
    def transfer_std_bkg_scaling(self, new_value):
        self.current_configuration.transfer_config.std_bkg_scaling = new_value
        self.update_transfer_function()

    @property
    def transfer_sample_pattern(self):
        """
        :rtype: Pattern
        """
        return self.current_configuration.transfer_config.sample_pattern

    @transfer_sample_pattern.setter
    def transfer_sample_pattern(self, new_pattern):
        self.current_configuration.transfer_config.sample_pattern = new_pattern

    @property
    def transfer_sample_bkg_pattern(self):
        """
        :rtype Pattern:
        """
        return self.current_configuration.transfer_config.sample_bkg_pattern

    @transfer_sample_bkg_pattern.setter
    def transfer_sample_bkg_pattern(self, new_pattern):
        self.current_configuration.transfer_config.sample_bkg_pattern = new_pattern
        self.update_transfer_function()

    @property
    def transfer_sample_bkg_scaling(self):
        return self.current_configuration.transfer_config.sample_bkg_scaling

    @transfer_sample_bkg_scaling.setter
    def transfer_sample_bkg_scaling(self, new_value):
        self.current_configuration.transfer_config.sample_bkg_scaling = new_value
        self.update_transfer_function()

    def set_smooth(self, value):
        self.original_pattern.set_smoothing(value)
        if self.background_pattern is not None:
            self.current_configuration.background_pattern.set_smoothing(value)
        self.calculate_transforms()

    def update_parameter(self,
                         sample_config: Sample,
                         transform_config: TransformConfiguration,
                         extrapolation_config: ExtrapolationConfiguration,
                         optimize_active,
                         r_cutoff,
                         optimize_iterations,
                         optimize_attenuation):

        self.auto_update = False
        self.sample = sample_config
        validate_composition(self.sample.composition, self.sample.sf_source)

        self.transform_config = transform_config
        self.extrapolation_config = extrapolation_config

        self.optimize = optimize_active
        self.r_cutoff = r_cutoff
        self.optimization_iterations = optimize_iterations
        self.optimization_attenuation = optimize_attenuation

        self.auto_update = True
        self.calculate_transforms()

    def calculate_transforms(self):
        if not self.auto_update:
            return

        if len(self.composition) != 0 and self.original_pattern:
            self.calculate_sq()

            if self.optimize:
                self.sq_pattern = optimize_sq(
                    self.sq_pattern, self.r_cutoff,
                    iterations=self.optimization_iterations,
                    atomic_density=convert_density_to_atoms_per_cubic_angstrom(
                        self.composition,
                        self.density),
                    use_modification_fcn=False,
                    attenuation_factor=self.optimization_attenuation,
                    fcn_callback=self.optimization_callback,
                    fourier_transform_method=self.transform_config.fourier_transform_method)
            self.calculate_fr()
            self.calculate_gr()
        self.data_changed.emit()

    def calculate_sq(self):
        if self.background_pattern is not None:
            sample_pattern = (self.original_pattern - self.background_pattern) \
                .limit(self.q_min, self.q_max)
        else:
            sample_pattern = self.original_pattern.limit(
                self.q_min, self.q_max)

        if self.use_transfer_function and self.transfer_function is not None:
            sample_pattern.y = sample_pattern.y * \
                               self.transfer_function(sample_pattern.x)

        if self.use_soller_correction:
            q, intensity = sample_pattern.data
            if self.soller_correction is None or \
                    self.soller_correction._max_thickness < self.soller_parameters['sample_thickness'] or \
                    self.soller_correction.wavelength != self.soller_parameters['wavelength'] or \
                    self.soller_correction._inner_radius != self.soller_parameters['inner_radius'] or \
                    self.soller_correction._outer_radius != self.soller_parameters['outer_radius'] or \
                    self.soller_correction._inner_width != self.soller_parameters['inner_width'] or \
                    self.soller_correction._outer_width != self.soller_parameters['outer_width'] or \
                    self.soller_correction._inner_length != self.soller_parameters['inner_length'] or \
                    self.soller_correction._outer_length != self.soller_parameters['outer_length']:

                if 2 > self.soller_parameters['sample_thickness']:
                    max_thickness = 2
                else:
                    max_thickness = self.soller_parameters["sample_thickness"] * 1.5

                self.soller_correction = SollerCorrectionGui(
                    q=q,
                    wavelength=self.soller_parameters['wavelength'],
                    max_thickness=max_thickness,
                    inner_radius=self.soller_parameters['inner_radius'],
                    outer_radius=self.soller_parameters['outer_radius'],
                    inner_width=self.soller_parameters['inner_width'],
                    outer_width=self.soller_parameters['outer_width'],
                    inner_length=self.soller_parameters['inner_length'],
                    outer_length=self.soller_parameters['outer_length'])

            sample_pattern = Pattern(
                q, self.soller_correction.transfer_function_sample(
                    self.soller_parameters['sample_thickness']) * intensity)

        self.sq_pattern = calculate_sq(
            sample_pattern,
            density=self.density,
            composition=self.composition,
            normalization_method=self.normalization_method,
            method=self.sq_method,
            sf_source=self.sf_source,
        )
        self.perform_extrapolation()

    def perform_extrapolation(self):
        if not self.extrapolation_config.activate:
            return

        extrapolation_method = self.extrapolation_config.method
        if self.extrapolation_config.s0_auto:
            self.extrapolation_config.s0 = calculate_s0(self.composition, self.sf_source)

        s0 = self.extrapolation_config.s0

        if extrapolation_method == 'step':
            self.sq_pattern = extrapolate_to_zero_step(self.sq_pattern, y0=s0)
        elif extrapolation_method == 'linear':
            self.sq_pattern = extrapolate_to_zero_linear(self.sq_pattern, y0=s0)
        elif extrapolation_method == 'spline':
            self.sq_pattern = extrapolate_to_zero_spline(
                self.sq_pattern,
                self.extrapolation_config.fit_q_max,
                y0=s0,
                replace=self.extrapolation_config.fit_replace)
        elif extrapolation_method == 'poly':
            self.sq_pattern = extrapolate_to_zero_poly(
                self.sq_pattern,
                self.extrapolation_config.fit_q_max,
                y0=s0,
                replace=self.extrapolation_config.fit_replace)

    def calculate_fr(self):
        self.fr_pattern = calculate_fr(
            self.sq_pattern,
            r=np.arange(self.r_min, self.r_max + self.r_step * 0.5, self.r_step),
            method=self.transform_config.fourier_transform_method,
            use_modification_fcn=self.use_modification_fcn)

    def calculate_gr(self):
        self.gr_pattern = calculate_gr(
            self.fr_pattern, self.density, self.composition)


    def optimize_density_and_scaling(self, density_min, density_max, bkg_min, bkg_max, iterations, callback_fcn=None,
                                     output_txt=None):
        params = Parameters()
        params.add("density", value=self.density, min=density_min, max=density_max)
        params.add("background_scaling", value=self.background_scaling, min=bkg_min, max=bkg_max)

        self.iteration = 0

        def optimization_fcn(params):
            density = params['density'].value
            background_scaling = params['background_scaling'].value

            self.background_pattern.scaling = background_scaling
            self.density = density
            self.calculate_transforms()
            # self.optimize_sq(iterations, fcn_callback=callback_fcn)

            r, gr = self.gr_pattern.limit(0, self.r_cutoff * 0.5).data

            output = gr ** 2

            self.write_output(u'{} X^2: {:.3f} Bkg_Scaling: {:.2f} Den: {:.3f}'.format(
                self.iteration, np.sum(output) / len(r), background_scaling, density),
                output_txt
            )
            self.iteration += 1
            return output

        res = minimize(optimization_fcn, params, method='least_squares', xtol=1e-3)
        self.write_fit_results(res.params, output_txt)
        return res.params

    def write_output(self, msg, output_txt=None):
        if output_txt is None:
            print(msg)
        else:
            print(msg)
            previous_txt = str(output_txt.toPlainText())
            new_txt = previous_txt + "\n" + str(msg)
            output_txt.setPlainText(new_txt)
            output_txt.verticalScrollBar().setValue(output_txt.verticalScrollBar().maximum())
            QtCore.QCoreApplication.processEvents()
            output_txt.verticalScrollBar().setValue(output_txt.verticalScrollBar().maximum())
            QtCore.QCoreApplication.processEvents()

    def write_fit_results(self, params, output_txt=None):
        output = '\nFit Results:\n'
        if params['density'].stderr is None:
            params['density'].stderr = np.nan
        if params['background_scaling'].stderr is None:
            params['background_scaling'].stderr = np.nan
        output += '-Background Scaling:\n  % .3g +/- %.3g\n' % ( \
            params['background_scaling'].value, params['background_scaling'].stderr)
        output += '-Density:\n  % .3g +/- %.3g\n' % (params['density'].value, params['density'].stderr)
        self.write_output(output, output_txt)

    def set_diamond_content(self, content_value):
        if content_value == 0:
            self.diamond_bkg_pattern = None
            self.calculate_transforms()
            return

        q, _ = self.background_pattern.data
        int = calculate_incoherent_scattering({'C': 1}, q) * content_value
        self.diamond_bkg_pattern = Pattern(q, int)
        self.calculate_transforms()

    def optimize_diamond_content(self, diamond_content=0, callback_fcn=None):
        params = Parameters()
        if diamond_content == 0:
            diamond_content = 20
        params.add("content", value=diamond_content, min=0)

        def optimization_fcn(params):
            diamond_content = params['content'].value
            self.set_diamond_content(diamond_content)
            low_r_pattern = self.gr_pattern.limit(0, self.r_cutoff)
            if callback_fcn is not None:
                callback_fcn(diamond_content)
            return low_r_pattern.data[1]

        result = minimize(optimization_fcn, params)
        print(result)

    def update_transfer_function(self):
        if self.transfer_std_pattern is None or \
                self.transfer_sample_pattern is None or \
                not self.use_transfer_function:
            return
        q_min = np.max([self.transfer_std_pattern.x[0], self.transfer_sample_pattern.x[0]])
        q_max = np.min([self.transfer_std_pattern.x[-1], self.transfer_sample_pattern.x[-1]])

        if self.transfer_std_bkg_pattern is None:
            std_pattern = self.transfer_std_pattern
        else:
            std_pattern = self.transfer_std_pattern - self.transfer_std_bkg_scaling * self.transfer_std_bkg_pattern

        if self.transfer_sample_bkg_pattern is None:
            sample_pattern = self.transfer_sample_pattern
        else:
            sample_pattern = self.transfer_sample_pattern - \
                             self.transfer_sample_bkg_scaling * \
                             self.transfer_sample_bkg_pattern

        self.current_configuration.transfer_config.function = \
            calculate_transfer_function(
                std_pattern.limit(q_min, q_max),
                sample_pattern.limit(q_min, q_max),
                smooth_factor=self.transfer_function_smoothing
            )
        self.calculate_transforms()

    def load_transfer_std_pattern(self, filename):
        self.transfer_std_pattern = Pattern.from_file(filename)

    def load_transfer_std_bkg_pattern(self, filename):
        self.transfer_std_bkg_pattern = Pattern.from_file(filename)

    def load_transfer_sample_pattern(self, filename):
        self.transfer_sample_pattern = Pattern.from_file(filename)

    def load_transfer_sample_bkg_pattern(self, filename):
        self.transfer_sample_bkg_pattern = Pattern.from_file(filename)


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.bool_):
            return bool(obj)
        if isinstance(obj, (np.floating, np.complexfloating)):
            return float(obj)
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.string_):
            return str(obj)
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        if isinstance(obj, timedelta):
            return str(obj)
        return super(NpEncoder, self).default(obj)


def validate_composition(composition: dict[str, float], sf_source: str):
    for element in list(composition.keys()):
        if element not in get_available_elements(sf_source):
            del composition[element]



# -*- coding: utf-8 -*-

from .composition import CompositionWidget
from .data import DataWidget
from .optimization import OptimizationWidget
from .options import OptionsWidget
from .density_optimization import DensityOptimizationWidget
from .extrapolation import ExtrapolationWidget
from .diamond import DiamondWidget
from .configuration import ConfigurationWidget
from .soller import SollerWidget
from .transfer import TransferFunctionWidget


# -*- coding: utf-8 -*-
from typing import Optional
from copy import deepcopy

import numpy as np
from scipy.integrate import simps

from . import calculate_gr_raw
from .scattering_factors import calculate_coherent_scattering_factor, calculate_incoherent_scattered_intensity, \
    ScatteringFactorCalculatorHajdu
from .soller_correction import SollerCorrection
from .pattern import Pattern

scattering_factor_param = ScatteringFactorCalculatorHajdu().coherent_param


def calculate_atomic_number_sum(composition: dict[str, float]):
    """
    Calculates the sum of the atomic number of all elements in the composition

    :param composition: composition as a dictionary with the elements as keys and the abundances as values
    :return: sum of the atomic numbers
    """
    z_tot = 0
    for element, n in composition.items():
        z_tot += scattering_factor_param['Z'][element] * n
    return z_tot


def calculate_effective_form_factors(composition: dict[str, float], q: np.ndarray) -> np.ndarray:
    """
    Calculates the effective form factor as defined in Eq. 10 in Eggert et al. (2002)

    :param composition: composition as a dictionary with the elements as keys and the abundances as values
    :param q: Q value or numpy array with a unit of A^-1
    :return: effective form factors numpy array
    """
    z_tot = calculate_atomic_number_sum(composition)

    f_effective = np.zeros_like(q)
    for element, n in composition.items():
        f_effective += calculate_coherent_scattering_factor(element, q) * n

    return f_effective / float(z_tot)


def calculate_incoherent_scattering(composition: dict[str, float], q: np.ndarray) -> np.ndarray:
    """
    Calculates the not normalized incoherent scattering contribution from a specific composition.

    :param composition:
    :param q: Q value or numpy array with a unit of A^-1
    :return: incoherent scattering numpy array
    """
    inc = np.zeros_like(q)
    for element, n in composition.items():
        inc += calculate_incoherent_scattered_intensity(element, q) * n

    return inc


def calculate_j(incoherent_scattering: np.ndarray, z_tot: float, f_effective: np.ndarray) -> np.ndarray:
    """
    Calculates the J parameter as described in equation (35) from Eggert et al. 2002.

    :param incoherent_scattering: Q dependent incoherent scattering
    :param z_tot: sum of atomic numbers for the material
    :param f_effective: Q dependent effective form factor
    :return: J numpy array with the same q as incoherent scattering and f_effective
    """
    return incoherent_scattering / (z_tot * f_effective) ** 2


def calculate_kp(element: str, f_effective: np.ndarray, q: np.ndarray) -> np.ndarray:
    """
    Calculates the average effective atomic number (averaged over the whole Q range).

    :param element: elemental symbol
    :param f_effective: effective form factor
    :param q: Q value or numpy array with a unit of A^-1
    :return: average effective atomic number
    :rtype: float
    """
    kp = np.mean(calculate_coherent_scattering_factor(element, q) / f_effective)
    return kp


def calculate_s_inf(composition: dict[str, float], z_tot: float, f_effective: np.ndarray, q: np.ndarray) -> float:
    """
    Calculates S_inf as described in equation (19) from Eggert et al. 2002

    :param composition: composition as a dictionary with the elements as keys and the abundances as values
    :param z_tot: sum of atomic numbers for the material
    :param f_effective: Q dependent effective form factor
    :param q: q numpy array with units of A^-1
    :return: S_inf value
    """
    sum_kp_squared = 0
    for element, n in composition.items():
        sum_kp_squared += n * calculate_kp(element, f_effective, q) ** 2

    return sum_kp_squared / z_tot ** 2


def calculate_alpha(sample_pattern: Pattern, z_tot: float, f_effective: np.ndarray, s_inf: np.ndarray, j: np.ndarray,
                    atomic_density: float) -> float:
    """
    Calculates the normalization factor alpha after equation (34) from Eggert et al. 2002.

    :param sample_pattern: Background subtracted sample pattern
    :param z_tot: sum opf atomic numbers for the material
    :param f_effective: Q dependent effective form factor
    :param s_inf: S_inf value (equ. (19) from Eggert et al. 2002)
    :param j: J value (equ. (35) from Eggert et al. 2002)
    :param atomic_density: number density in atoms/Angstrom^3
    :return: normalization factor alpha
    """

    q, intensity = sample_pattern.data

    integral_1 = simps((j + s_inf) * q ** 2, q)
    integral_2 = simps((intensity / f_effective ** 2) * q ** 2, q)

    alpha = z_tot ** 2 * (-2 * np.pi ** 2 * atomic_density + integral_1) / integral_2

    return alpha


def calculate_coherent_scattering(sample_pattern: Pattern, alpha: float, n: float,
                                  incoherent_scattering: np.ndarray) -> Pattern:
    """
    Calculates the coherent Scattering Intensity Pattern

    :param sample_pattern:  Background subtracted sample pattern
    :param alpha: normalization factor alpha (after equ. (34) from Eggert et al. 2002)
    :param n: Number of atoms
    :param incoherent_scattering: incoherent scattering intensity
    :return: Coherent Scattering Pattern
    :rtype: Pattern
    """

    q, intensity = sample_pattern.data
    coherent_intensity = n * (alpha * intensity - incoherent_scattering)
    return Pattern(q, coherent_intensity)


def calculate_sq(coherent_pattern: Pattern, n: float, z_tot: float, f_effective: np.ndarray) -> Pattern:
    """
    Calculates the Structure Factor based on equation (18) in Eggert et al. 2002

    :param coherent_pattern: coherent pattern
    :param n: number of atoms for structural unit, e.g. 3 for SiO2
    :param z_tot: sum of atomic numbers for the material
    :param f_effective: Q dependent effective form factor
    :return: S(q)
    """
    q, coherent_intensity = coherent_pattern.data
    sq_intensity = coherent_intensity / (n * z_tot ** 2 * f_effective ** 2)

    return Pattern(q, sq_intensity)


def calculate_fr(iq_pattern: Pattern, r: Optional[np.ndarray] = None, use_modification_fcn: bool = False) -> Pattern:
    """
    Calculates F(r) from a given interference function i(Q) for r values.
    If r is none a range from 0 to 10 with step 0.01 is used.    A Lorch modification function of the form:

        m = sin(q*pi/q_max)/(q*pi/q_max)

    can be used to address issues with a low q_max. This will broaden the sharp peaks in f(r)

    :param iq_pattern:             interference function i(q) = S(Q)-S_inf with lim_inf i(Q)=0 and unit(q)=A^-1
    :type iq_pattern: Pattern
    :param r:                       numpy array giving the r-values for which F(r) will be calculated,
                                    default is 0 to 10 with 0.01 as a step. units should be in Angstrom.
    :param use_modification_fcn:    boolean flag whether to use the Lorch modification function

    :return: F(r) pattern
    :rtype: Pattern
    """
    if r is None:
        r = np.arange(0, 10, 0.01)

    q, iq = iq_pattern.data
    if use_modification_fcn:
        modification = np.sin(q * np.pi / np.max(q)) / (q * np.pi / np.max(q))
    else:
        modification = 1

    fr = 2.0 / np.pi * simps(modification * q * iq * np.array(np.sin(np.outer(q.T, r))).T, q)

    return Pattern(r, fr)


def optimize_iq(iq_pattern, r_cutoff, iterations, atomic_density, j, s_inf=1, use_modification_fcn=False,
                attenuation_factor=1, fcn_callback=None, callback_period=2):
    """
    Performs an optimization of the structure factor based on an r_cutoff value as described in Eggert et al. 2002 PRB,
    65, 174105. This basically does back and forward transforms between S(Q) and f(r) until the region below the
    r_cutoff value is a flat line without any oscillations.

    :param iq_pattern:
        original i(Q) pattern = S(Q)-S_inf
    :param r_cutoff:
        cutoff value below which there is no signal expected (below the first peak in g(r))
    :param iterations:
        number of back and forward transforms
    :param atomic_density:
        density in atoms/A^3
    :param j:
        J value (equ. (35) from Eggert et al. 2002)
    :param s_inf:
        S_inf value (equ. (19) from Eggert et al. 2002, defaults to 1, which is the value for mon-atomic substances
    :param use_modification_fcn:
        Whether to use the Lorch modification function during the Fourier transform.
        Warning: When using the Lorch modification function usually more iterations are needed to get to the
        wanted result.
    :param attenuation_factor:
        Sometimes the initial change during back and forward transformations results in a runaway, by setting the
        attenuation factor to higher than one can help for this situation, it basically reduces
        the amount of change during each iteration.
    :param fcn_callback:
        Function which will be called at an iteration period defined by the callback_period parameter.
        The function should take 3 arguments: sq_pattern, fr_pattern and gr_pattern. Additionally the function
        should return a boolean value, where True continues the optimization and False will stop the optimization
        procedure
    :param callback_period:
        determines how frequently the fcn_callback will be called.

    :return:
        optimized S(Q) pattern
    """
    r = np.arange(0, r_cutoff, 0.02)
    iq_pattern = deepcopy(iq_pattern)
    for iteration in range(iterations):
        fr_pattern = calculate_fr(iq_pattern, r, use_modification_fcn)
        q, iq_int = iq_pattern.data
        r, fr_int = fr_pattern.data

        delta_fr = fr_int + 4 * np.pi * r * atomic_density

        in_integral = np.array(np.sin(np.outer(q.T, r))) * delta_fr
        integral = np.trapz(in_integral, r) / attenuation_factor
        iq_optimized = iq_int - 1. / q * (iq_int / (s_inf + j) + 1) * integral

        iq_pattern = Pattern(q, iq_optimized)

        if fcn_callback is not None and iteration % callback_period == 0:
            fr_pattern = calculate_fr(iq_pattern, use_modification_fcn=use_modification_fcn)
            gr_pattern = calculate_gr_raw(fr_pattern, atomic_density)
            fcn_callback(iq_pattern, fr_pattern, gr_pattern)
    return iq_pattern


def calculate_chi2_map(data_pattern, bkg_pattern, composition,
                       densities, bkg_scalings, r_cutoff, iterations=2):
    """
    Calculates a chi2 2d array for an array of densities and background scalings.

    :param data_pattern: original data pattern
    :param bkg_pattern: original background pattern
    :param composition: composition as a dictionary with the elements as keys and the abundances as values
    :param densities: 1-dimensional array of densities for which to calculate chi2
    :param bkg_scalings: 1-dimensional array of background scalings for which to calculate chi2
    :param r_cutoff: cutoff value below which there is no signal expected (below the first peak in g(r))
    :param iterations: number of iterations for optimization, described in equations 47-49 in Eggert et al. 2002
    :return: 2-dimensional array of chi2 values
    """

    n = sum([composition[x] for x in composition])
    q = data_pattern.extend_to(0, 0).x

    inc = calculate_incoherent_scattering(composition, q)
    f_eff = calculate_effective_form_factors(composition, q)
    z_tot = calculate_atomic_number_sum(composition)
    s_inf = calculate_s_inf(composition, z_tot, f_eff, q)
    j = calculate_j(inc, z_tot, f_eff)

    chi2 = np.zeros((len(densities), len(bkg_scalings)))

    for n1, density in enumerate(densities):
        for n2, bkg_scaling in enumerate(bkg_scalings):
            # density = params['density'].value
            # bkg_scaling = params['bkg_scaling'].value

            r = np.arange(0, r_cutoff, 0.02)
            sample_pattern = data_pattern - bkg_scaling * bkg_pattern
            sample_pattern = sample_pattern.extend_to(0, 0)

            alpha = calculate_alpha(sample_pattern, z_tot, f_eff, s_inf, j, density)

            coherent_pattern = calculate_coherent_scattering(sample_pattern, alpha, n, inc)
            sq_pattern = calculate_sq(coherent_pattern, n, z_tot, f_eff)
            iq_pattern = Pattern(sq_pattern.x, sq_pattern.y - s_inf)

            delta_fr = np.zeros(r.shape)

            for iteration in range(iterations):
                fr_pattern = calculate_fr(iq_pattern, r)

                q, iq_int = iq_pattern.data
                r, fr_int = fr_pattern.data

                delta_fr = fr_int + 4 * np.pi * r * density

                in_integral = np.array(np.sin(np.outer(q.T, r))) * delta_fr
                integral = np.trapz(in_integral, r)
                iq_optimized = iq_int - 1. / q * (iq_int / (s_inf + j) + 1) * integral

                iq_pattern = Pattern(q, iq_optimized)

            chi2[n1, n2] = np.sum(delta_fr ** 2)
    return chi2


def optimize_density_and_bkg_scaling(data_pattern: Pattern, bkg_pattern: Pattern, composition: dict[str, float],
                                     initial_density: float, initial_bkg_scaling: float, r_cutoff: float,
                                     iterations: int = 2, use_modification_fcn: bool = False) \
        -> tuple[float, float, float, float]:
    """
    This function tries to find the optimum density in background scaling with the given parameters. The equations
    behind the optimization are presented equ (47-50) in the Eggert et al. 2002 paper.

    :param data_pattern: original data pattern
    :param bkg_pattern: original background pattern
    :param composition: composition as a dictionary with the elements as keys and the abundances as values
    :param initial_density: density starting point for the optimization procedure
    :param initial_bkg_scaling: background scaling starting point for the optimization procedure
    :param r_cutoff: cutoff value below which there is no signal expected (below the first peak in g(r))
    :param iterations: number of iterations for optimization, described in equations 47-49 in Eggert et al. 2002
    :param use_modification_fcn: Whether or not to use the Lorch modification function during the Fourier transform.
    :return: tuple with optimized parameters (density, density_error, bkg_scaling, bkg_scaling_error)
    """

    N = sum([composition[x] for x in composition])
    q = data_pattern.extend_to(0, 0).x

    inc = calculate_incoherent_scattering(composition, q)
    f_eff = calculate_effective_form_factors(composition, q)
    z_tot = calculate_atomic_number_sum(composition)
    s_inf = calculate_s_inf(composition, z_tot, f_eff, q)
    j = calculate_j(inc, z_tot, f_eff)

    def optimization_fcn(x):
        density = x['density'].value
        bkg_scaling = x['bkg_scaling'].value

        r = np.arange(0, r_cutoff, 0.02)
        sample_pattern = data_pattern - bkg_scaling * bkg_pattern
        sample_pattern = sample_pattern.extend_to(0, 0)

        alpha = calculate_alpha(sample_pattern, z_tot, f_eff, s_inf, j, density)

        coherent_pattern = calculate_coherent_scattering(sample_pattern, alpha, N, inc)
        sq_pattern = calculate_sq(coherent_pattern, N, z_tot, f_eff)
        iq_pattern = Pattern(sq_pattern.x, sq_pattern.y - s_inf)

        fr_pattern = calculate_fr(iq_pattern, r, use_modification_fcn=use_modification_fcn)

        q, iq_int = iq_pattern.data
        r, fr_int = fr_pattern.data

        delta_fr = fr_int + 4 * np.pi * r * density

        for iteration in range(iterations):
            in_integral = np.array(np.sin(np.outer(q.T, r))) * delta_fr
            integral = np.trapz(in_integral, r)
            iq_optimized = iq_int - 1. / q * (iq_int / (s_inf + j) + 1) * integral

            iq_pattern = Pattern(q, iq_optimized)
            fr_pattern = calculate_fr(iq_pattern, r)

            q, iq_int = iq_pattern.data
            r, fr_int = fr_pattern.data

            delta_fr = fr_int + 4 * np.pi * r * density

        return delta_fr

    from lmfit import Parameters, minimize

    params = Parameters()
    params.add('density', value=initial_density, )
    params.add('bkg_scaling', value=initial_bkg_scaling)

    result = minimize(optimization_fcn, params)

    return result.params['density'].value, result.params['density'].stderr, \
        result.params['bkg_scaling'].value, result.params['density'].stderr


def optimize_soller_dac(data_pattern: Pattern, bkg_pattern: Pattern, composition: dict[str, float],
                        initial_density: float, initial_bkg_scaling: float, initial_thickness: float,
                        sample_thickness: float, wavelength: float, initial_carbon_content: float = 1,
                        r_cutoff: float = 2.28, iterations: int = 1, use_modification_fcn: bool = False,
                        vary: tuple[bool, bool, bool] = (True, True, True)) \
        -> tuple[float, float, float, float, float, float, float]:
    """
    Optimizes density, background scaling and diamond content for a list of sample thickness with a given initial
    gasket thickness in the diamond anvil cell (DAC). The calculation is done by utilizing the soller slit transfer
    function and assuming that the DAC has been centered to the rotation center of the soller slit.

    :param data_pattern: original data pattern
    :param bkg_pattern: original background pattern
    :param composition: composition as a dictionary with the elements as keys and the abundances as values
    :param initial_density: density starting point for the optimization procedure
    :param initial_bkg_scaling: background scaling starting point for the optimization procedure
    :param initial_thickness: gasket thickness with which the background was measured.
    :param sample_thickness: sample thickness for which the sample was measured
    :param wavelength: wavelength of the radiation used - needed for calculation of soller slit transfer function in
                       q-space
    :param initial_carbon_content: carbon content starting point for the optimization
    :param r_cutoff: cutoff value below which there is no signal expected (below the first peak in g(r)
    :param iterations: number of iterations for optimization, described in equations 47-49 in Eggert et al. 2002
    :param use_modification_fcn: Whether or not to use the Lorch modification function during the Fourier transform.
    :param vary: 3 boolean flags whether to vary: density, bkg_scaling, carbon_content during the optimization
    :return:
    """
    n = sum([composition[x] for x in composition])
    q = data_pattern.extend_to(0, 0).x

    inc = calculate_incoherent_scattering(composition, q)
    f_eff = calculate_effective_form_factors(composition, q)
    z_tot = calculate_atomic_number_sum(composition)
    s_inf = calculate_s_inf(composition, z_tot, f_eff, q)
    j = calculate_j(inc, z_tot, f_eff)

    tth = 2 * np.arcsin(data_pattern.x * wavelength / (4 * np.pi)) / np.pi * 180
    soller = SollerCorrection(tth, initial_thickness)

    def optimization_fcn(params):
        diamond_content = params['diamond_content'].value
        bkg_scaling = params['bkg_scaling'].value
        density = params['density'].value

        q, data_int = data_pattern.data
        _, bkg_int = bkg_pattern.data

        sample_transfer, diamond_transfer = soller.transfer_function_dac(sample_thickness, initial_thickness)

        diamond_background = diamond_content * Pattern(q,
                                                       calculate_incoherent_scattering({'C': 1},
                                                                                       q) / diamond_transfer)

        sample_pattern = data_pattern - bkg_scaling * bkg_pattern
        sample_pattern = sample_pattern - diamond_background
        sample_pattern = Pattern(q, sample_pattern.y * sample_transfer)
        sample_pattern = sample_pattern.extend_to(0, 0)

        alpha = calculate_alpha(sample_pattern, z_tot, f_eff, s_inf, j, density)

        coherent_pattern = calculate_coherent_scattering(sample_pattern, alpha, n, inc)
        sq_pattern = calculate_sq(coherent_pattern, n, z_tot, f_eff)
        iq_pattern = Pattern(sq_pattern.x, sq_pattern.y - s_inf)

        r = np.arange(0, r_cutoff, 0.02)
        fr_pattern = calculate_fr(iq_pattern, r, use_modification_fcn=use_modification_fcn)

        q, iq_int = iq_pattern.data
        r, fr_int = fr_pattern.data

        delta_fr = fr_int + 4 * np.pi * r * density

        for iteration in range(iterations):
            in_integral = np.array(np.sin(np.outer(q.T, r))) * delta_fr
            integral = np.trapz(in_integral, r)
            iq_optimized = iq_int - 1. / q * (iq_int / (s_inf + j) + 1) * integral

            iq_pattern = Pattern(q, iq_optimized)
            fr_pattern = calculate_fr(iq_pattern, r)

            q, iq_int = iq_pattern.data
            r, fr_int = fr_pattern.data

            delta_fr = fr_int + 4 * np.pi * r * density

        return delta_fr

    from lmfit import Parameters, minimize, report_fit

    params = Parameters()
    params.add('density', value=initial_density, min=0, vary=vary[0])
    params.add('bkg_scaling', value=initial_bkg_scaling, vary=vary[1])
    params.add('diamond_content', value=initial_carbon_content, min=0, vary=vary[2])

    result = minimize(optimization_fcn, params)

    report_fit(result)

    return result.chisqr, \
        result.params['density'].value, result.params['density'].stderr, \
        result.params['bkg_scaling'].value, result.params['bkg_scaling'].stderr, \
        result.params['diamond_content'].value, result.params['diamond_content'].stderr



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import sys

from glassure import __version__
from glassure.gui.controller.glassure_controller import GlassureController
from qtpy import QtWidgets


def my_exception_hook(exctype, value, traceback):
    # Print the error and traceback
    print(exctype, value, traceback)


# sys.excepthook = my_exception_hook


def main():
    app = QtWidgets.QApplication(sys.argv)
    from sys import platform as _platform

    print("Glassure {}".format(__version__)) 

    if _platform != "Darwin":
        app.setStyle('plastique')
    controller = GlassureController()
    controller.show_window()
    app.exec_()
    del app


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from typing import Optional
import numpy as np
import lmfit

from . import Pattern
from .utility import calculate_incoherent_scattering, calculate_f_squared_mean, calculate_f_mean_squared, \
    convert_density_to_atoms_per_cubic_angstrom

from .methods import SqMethod, NormalizationMethod, FourierTransformMethod

__all__ = ['calculate_normalization_factor_raw', 'calculate_normalization_factor', 'fit_normalization_factor',
           'calculate_sq', 'calculate_sq_raw', 'calculate_sq_from_fr', 'calculate_sq_from_gr',
           'calculate_fr', 'calculate_gr_raw', 'calculate_gr']


def calculate_normalization_factor_raw(sample_pattern: Pattern, atomic_density: float, f_squared_mean: np.ndarray,
                                       f_mean_squared: np.ndarray, incoherent_scattering: Optional[np.ndarray] = None,
                                       attenuation_factor: float = 0.001) -> float:
    """
    Calculates the normalization factor for a sample pattern given all the parameters. If you do not have them
    already calculated please consider using calculate_normalization_factor, which has an easier interface since it
    just requires density and composition as parameters.

    :param sample_pattern:     background subtracted sample pattern
    :param atomic_density:      density in atoms per cubic Angstrom
    :param f_squared_mean:      <f^2>
    :param f_mean_squared:      <f>^2
    :param incoherent_scattering: compton scattering from sample, if set to None, it will not be used
    :param attenuation_factor:  attenuation factor used in the exponential, in order to correct for the q cutoff

    :return:                    normalization factor
    """
    q, intensity = sample_pattern.data
    # calculate values for integrals
    if incoherent_scattering is None:
        incoherent_scattering = np.zeros_like(q)
    n1 = q ** 2 * ((f_squared_mean + incoherent_scattering) * np.exp(-attenuation_factor * q ** 2)) / f_mean_squared
    n2 = q ** 2 * intensity * np.exp(-attenuation_factor * q ** 2) / f_mean_squared

    n = ((-2 * np.pi ** 2 * atomic_density + np.trapz(q, n1)) / np.trapz(q, n2))

    return n


def calculate_normalization_factor(sample_pattern: Pattern, density: float, composition: dict[str, float],
                                   attenuation_factor: float = 0.001, use_incoherent_scattering: bool = True) -> float:
    """
    Calculates the normalization factor for a background subtracted sample pattern based on density and composition.

    :param sample_pattern:     background subtracted sample pattern with A-1 as x unit
    :param density:             density in g/cm^3
    :param composition:         composition as a dictionary with the elements as keys and the abundances as values
    :param attenuation_factor:  attenuation factor used in the exponential, in order to correct for the q cutoff
    :param use_incoherent_scattering: whether to use incoherent scattering, in some cases it is already subtracted

    :return: normalization factor
    """
    q, intensity = sample_pattern.data

    f_squared_mean = calculate_f_squared_mean(composition, q)
    f_mean_squared = calculate_f_mean_squared(composition, q)
    if use_incoherent_scattering:
        incoherent_scattering = calculate_incoherent_scattering(composition, q)
    else:
        incoherent_scattering = None
    atomic_density = convert_density_to_atoms_per_cubic_angstrom(composition, density)

    return calculate_normalization_factor_raw(sample_pattern, atomic_density, f_squared_mean, f_mean_squared,
                                              incoherent_scattering, attenuation_factor)


def fit_normalization_factor(sample_pattern: Pattern, composition: dict[str, float], q_cutoff: float = 3,
                             method: str = "linear", use_incoherent_scattering: bool = True,
                             sf_source: str = 'hajdu') -> float:
    """
    Estimates the normalization factor n for calculating S(Q) by fitting
        (Intensity*n-Multiple Scattering) * Q^2
    to
        (Incoherent Scattering + Self Scattering) * Q^2
    where n and Multiple Scattering are free parameters.

    :param sample_pattern:      background subtracted sample pattern with A^-1 as x unit
    :param composition:         composition as a dictionary with the elements as keys and the abundances as values
    :param q_cutoff:            q value above which the fitting will be performed, default = 3
    :param method:              specifies whether q^2 ("squared") or q (linear) should be used
    :param use_incoherent_scattering:
                                whether to use incoherent scattering, in some cases it is already subtracted
    :param sf_source:           source of the scattering factors. Possible sources are 'hajdu' and 'brown_hubbell'.

    :return: normalization factor
    """
    q, intensity = sample_pattern.limit(q_cutoff, 100000).data

    if method == "squared":
        x = q ** 2
    elif method == "linear":
        x = q
    else:
        raise NotImplementedError("{} is not an allowed method for fit_normalization_factor".format(method))

    theory = calculate_f_squared_mean(composition, q, sf_source) * x
    if use_incoherent_scattering:
        theory += x * calculate_incoherent_scattering(composition, q, sf_source)

    params = lmfit.Parameters()
    params.add("n", value=1, min=0)
    params.add("multiple", value=0, min=0)

    def optimization_fcn(params, x, sample_intensity, theory_intensity):
        n = params['n'].value
        multiple = params['multiple'].value
        return ((sample_intensity * n - multiple) * x - theory_intensity) ** 2

    out = lmfit.minimize(optimization_fcn, params, args=(x, intensity, theory))
    return out.params['n'].value


def calculate_sq_raw(sample_pattern: Pattern, f_squared_mean: np.ndarray, f_mean_squared: np.ndarray,
                     incoherent_scattering: Optional[np.ndarray] = None, normalization_factor: float = 1,
                     method: str = 'FZ') -> Pattern:
    """
    Calculates the structure factor of a material with the given parameters. Using the equation:

    S(Q) = (n * Intensity - incoherent_scattering - <f>^2-)/<f^2> + 1

    where n is the normalization factor and f are the scattering factors.

    :param sample_pattern:       background subtracted sample pattern with A^-1 as x unit
    :param f_squared_mean:        <f^2>
    :param f_mean_squared:        <f>^2
    :param incoherent_scattering: compton scattering from sample
    :param normalization_factor:  previously calculated normalization factor, if None, it will not be subtracted
    :param method:                describing the method to calculate the structure factor, possible values are
                                    - 'AL' - Ashcroft-Langreth
                                    - 'FZ' - Faber-Ziman

    :return: S(Q) pattern
    """
    q, intensity = sample_pattern.data
    if incoherent_scattering is None:
        incoherent_scattering = np.zeros_like(q)

    if method == 'FZ' or method == SqMethod.FZ:
        sq = (normalization_factor * intensity - incoherent_scattering - f_squared_mean + f_mean_squared) / \
             f_mean_squared
    elif method == 'AL' or method == SqMethod.AL:
        sq = (normalization_factor * intensity - incoherent_scattering) / f_squared_mean
    else:
        raise NotImplementedError('{} method is not implemented'.format(method))
    return Pattern(q, sq)


def calculate_sq(sample_pattern: Pattern, density: float, composition: dict[str, float],
                 attenuation_factor: float = 0.001, method: str = 'FZ',
                 normalization_method: str = 'int', use_incoherent_scattering: bool = True,
                 sf_source: str = 'hajdu') -> Pattern:
    """
    Calculates the structure factor of a material with the given parameters. Using the equation:

    S(Q) = (n * Intensity - incoherent_scattering - <f>^2-)/<f^2> + 1

    where n is the normalization factor and f are the scattering factors. All parameters from the equation are
    calculated from the density, composition and the sample pattern

    :param sample_pattern:      background subtracted sample pattern with A^-1 as x unit
    :param density:             density of the sample in g/cm^3
    :param composition:         composition as a dictionary with the elements as keys and the abundances as values
    :param attenuation_factor:  attenuation factor used in the exponential for the calculation of the normalization
                                factor

    :param method:              describing the method to calculate the structure factor, possible values are
                                    - 'AL' - Ashcroft-Langreth
                                    - 'FZ' - Faber-Ziman

    :param normalization_method: determines the method used for estimating the normalization method. possible values are
                                'int' for an integral or 'fit' for fitting the high q region form factors.

    :param use_incoherent_scattering: whether to use incoherent scattering, in some cases it is already subtracted
    :param sf_source:           source of the scattering factors. Possible sources are 'hajdu' and 'brown_hubbell'.

    :return: S(Q) pattern
    """
    q, intensity = sample_pattern.data
    f_squared_mean = calculate_f_squared_mean(composition, q, sf_source)
    f_mean_squared = calculate_f_mean_squared(composition, q, sf_source)
    if use_incoherent_scattering:
        incoherent_scattering = calculate_incoherent_scattering(composition, q, sf_source)
    else:
        incoherent_scattering = None

    atomic_density = convert_density_to_atoms_per_cubic_angstrom(composition, density)
    if normalization_method == 'fit' or normalization_method == NormalizationMethod.FIT:
        normalization_factor = fit_normalization_factor(sample_pattern,
                                                        composition,
                                                        use_incoherent_scattering,
                                                        sf_source=sf_source)
    else:
        normalization_factor = calculate_normalization_factor_raw(sample_pattern,
                                                                  atomic_density,
                                                                  f_squared_mean,
                                                                  f_mean_squared,
                                                                  incoherent_scattering,
                                                                  attenuation_factor)
    return calculate_sq_raw(sample_pattern,
                            f_squared_mean,
                            f_mean_squared,
                            incoherent_scattering,
                            normalization_factor,
                            method)


def calculate_fr(sq_pattern: Pattern, r: Optional[np.ndarray] = None, use_modification_fcn: bool = False,
                 method: str = 'integral') -> Pattern:
    """
    Calculates F(r) from a given S(Q) pattern for r values.
    If r is None, a range from 0 to 10 with step 0.01 is used.
    A Lorch modification function of the form:

        m = sin(q*pi/q_max)/(q*pi/q_max)

    can be used to address issues with a low q_max. This will broaden the sharp peaks in g(r)

    :param sq_pattern:              Structure factor S(Q) with lim_inf S(Q) = 1 and unit(q)=A^-1
    :param r:                       numpy array giving the r-values for which F(r) will be calculated,
                                    default is 0 to 10 with 0.01 as a step. units should be in Angstrom.
    :param use_modification_fcn:    boolean flag whether to use the Lorch modification function
    :param method:                  determines the method used for calculating fr, possible values are:
                                            - 'integral' solves the Fourier integral, by calculating the integral
                                            - 'fft' solves the Fourier integral by using fast fourier transformation

    :return: F(r) pattern
    """
    if r is None:
        r = np.linspace(0.01, 10, 1000)

    q, sq = sq_pattern.data
    if use_modification_fcn:
        modification = np.sin(q * np.pi / np.max(q)) / (q * np.pi / np.max(q))
    else:
        modification = 1

    if method == 'integral' or method == FourierTransformMethod.INTEGRAL:
        fr = 2.0 / np.pi * np.trapz(modification * q * (sq - 1) * np.array(np.sin(np.outer(q.T, r))).T, q)
    elif method == 'fft' or method == FourierTransformMethod.FFT:
        q_step = q[1] - q[0]
        r_step = r[1] - r[0]

        n_out = np.max([len(q), int(np.pi / (r_step * q_step))])
        q_max_for_ifft = 2 * n_out * q_step
        y_for_ifft = np.concatenate((modification * q * (sq - 1), np.zeros(2 * n_out - len(q))))

        ifft_result = np.fft.ifft(y_for_ifft) * 2 / np.pi * q_max_for_ifft
        ifft_imag = np.imag(ifft_result)[:n_out]
        ifft_x_step = 2 * np.pi / q_max_for_ifft
        ifft_x = np.arange(n_out) * ifft_x_step

        fr = np.interp(r, ifft_x, ifft_imag)
    else:
        raise NotImplementedError("{} is not an allowed method for calculate_fr".format(method))
    return Pattern(r, fr)


def calculate_sq_from_fr(fr_pattern: Pattern, q: np.ndarray, method: str = 'integral') -> Pattern:
    """
    Calculates S(Q) from an F(r) pattern for given q values.

    :param fr_pattern:              input F(r) pattern
    :param q:                       numpy array giving the q-values for which S(q) will be calculated,
    :param method:                  determines the method use for calculating fr, possible values are:
                                            - 'integral' solves the Fourier integral, by calculating the integral
                                            - 'fft' solves the Fourier integral by using fast fourier transformation

    :return: F(r) pattern
    """
    r, fr = fr_pattern.data

    if method == 'integral':
        sq = np.trapz(fr * np.array(np.sin(np.outer(r.T, q))).T, r) / q + 1

    elif method == 'fft':
        q_step = q[1] - q[0]
        r_step = r[1] - r[0]

        n_out = int(np.pi / (r_step * q_step))

        r_max_for_ifft = 2 * n_out * r_step
        ifft_x_step = 2 * np.pi / r_max_for_ifft
        ifft_x = np.arange(n_out) * ifft_x_step

        y_for_ifft = np.concatenate((fr, np.zeros(2 * n_out - len(r))))
        ifft_result = np.fft.ifft(y_for_ifft) * r_max_for_ifft
        ifft_imag = np.imag(ifft_result)[:n_out]

        sq = np.interp(q, ifft_x, ifft_imag) / q + 1
    else:
        raise NotImplementedError("{} is not an allowed method for calculate_sq_from_fr".format(method))

    return Pattern(q, sq)


def calculate_sq_from_gr(gr_pattern: Pattern, q: np.ndarray, density: float, composition: dict[str, float],
                         method: str = 'integral') -> Pattern:
    """
    Performs a back Fourier transform from the pair distribution function g(r)

    :param gr_pattern:      g(r) pattern
    :param q:               numpy array of q values for which S(Q) should be calculated
    :param density:         density of the sample in g/cm^3
    :param composition:     composition as a dictionary with the elements as keys and the abundances as values
    :param method:          determines the method used for calculating fr, possible values are:
                                - 'integral' solves the Fourier integral, by calculating the integral
                                - 'fft' solves the Fourier integral by using fast fourier transformation

    :return: S(Q) pattern
    """
    atomic_density = convert_density_to_atoms_per_cubic_angstrom(composition, density)
    r, gr = gr_pattern.data

    # removing the nan value at the first index, which is caused by the division by zero when r started from zero
    if np.isnan(gr[0]):
        gr[0] = 0
    fr_pattern = Pattern(r, (gr - 1) * (4.0 * np.pi * r * atomic_density))
    return calculate_sq_from_fr(fr_pattern, q, method)


def calculate_gr_raw(fr_pattern: Pattern, atomic_density: float) -> Pattern:
    """
    Calculates a g(r) pattern from a given F(r) pattern and the atomic density

    :param fr_pattern:     F(r) pattern
    :param atomic_density:  atomic density in atoms/A^3

    :return: g(r) pattern
    """
    r, f_r = fr_pattern.data
    g_r = 1 + f_r / (4.0 * np.pi * r * atomic_density)
    return Pattern(r, g_r)


def calculate_gr(fr_pattern: Pattern, density: float, composition: dict[str, float]) -> Pattern:
    """
    Calculates a g(r) pattern from a given F(r) pattern, the material density and composition.

    :param fr_pattern:     F(r) pattern
    :param density:         density in g/cm^3
    :param composition:     composition as a dictionary with the elements as keys and the abundances as values

    :return: g(r) pattern
    """
    return calculate_gr_raw(fr_pattern, convert_density_to_atoms_per_cubic_angstrom(composition, density))



# -*- coding: utf-8 -*-
from qtpy import QtWidgets


def open_file_dialog(parent_widget, caption, directory, filter=None):
    filename = QtWidgets.QFileDialog.getOpenFileName(parent_widget, caption=caption,
                                                     dir=directory,
                                                     filter=filter)
    if isinstance(filename, tuple):  # PyQt5 returns a tuple...
        return str(filename[0])
    return str(filename)


def open_files_dialog(parent_widget, caption, directory, filter=None):
    filenames = QtWidgets.QFileDialog.getOpenFileNames(parent_widget, caption=caption,
                                                       dir=directory,
                                                       filter=filter)
    if isinstance(filenames, tuple):  # PyQt5 returns a tuple...
        filenames = filenames[0]
    return filenames


def save_file_dialog(parent_widget, caption, directory, filter=None):
    filename = QtWidgets.QFileDialog.getSaveFileName(parent_widget, caption,
                                                     dir=directory,
                                                     filter=filter)
    if isinstance(filename, tuple):  # PyQt5 returns a tuple...
        return str(filename[0])
    return str(filename)



# -*- coding: utf-8 -*-

from copy import deepcopy

import numpy as np
import lmfit

from . import Pattern
from .calc import calculate_fr, calculate_gr_raw, calculate_sq, calculate_sq_raw, calculate_normalization_factor_raw, \
    fit_normalization_factor
from .utility import convert_density_to_atoms_per_cubic_angstrom, calculate_incoherent_scattering, \
    calculate_f_mean_squared, calculate_f_squared_mean
from .utility import extrapolate_to_zero_poly
from .soller_correction import SollerCorrection

__all__ = ['optimize_sq', 'optimize_density', 'optimize_incoherent_container_scattering',
           'optimize_soller_dac']


def optimize_sq(sq_pattern: Pattern, r_cutoff: float, iterations: int, atomic_density: float,
                use_modification_fcn: bool = False, attenuation_factor: float = 1,
                fcn_callback=None, callback_period: int = 2, fourier_transform_method: str = 'fft'):
    """
    Performs an optimization of the structure factor based on an r_cutoff value as described in Eggert et al. 2002 PRB,
    65, 174105. This basically does back and forward transforms between S(Q) and f(r) until the region below the
    r_cutoff value is a flat line without any oscillations.

    :param sq_pattern:
        original S(Q)
    :param r_cutoff:
        cutoff value below which there is no signal expected (below the first peak in g(r))
    :param iterations:
        number of back and forward transforms
    :param atomic_density:
        density in atoms/A^3
    :param use_modification_fcn:
        Whether to use the Lorch modification function during the Fourier transform.
        Warning: When using the Lorch modification function, usually more iterations are needed to get to the
        wanted result.
    :param attenuation_factor:
        Sometimes the initial change during back and forward transformations results in a runaway, by setting the
        attenuation factor reduces the amount of change during each iteration.
    :param fcn_callback:
        Function which will be called at an iteration period defined by the callback_period parameter.
        The function should take three arguments: sq_pattern, fr_pattern and gr_pattern.
        Additionally, the function should return a boolean value, where True continues the optimization and False will
        stop the optimization.
    :param callback_period:
        determines how frequently the fcn_callback will be called.
    :param fourier_transform_method:
        determines which method will be used for the Fourier transform. Possible values are 'fft' and 'integral'

    :return:
        optimized S(Q) pattern
    """
    r = np.arange(0, r_cutoff, 0.02)
    sq_pattern = deepcopy(sq_pattern)
    for iteration in range(iterations):
        fr_pattern = calculate_fr(sq_pattern, r, use_modification_fcn, method=fourier_transform_method)
        q, sq_int = sq_pattern.data
        r, fr_int = fr_pattern.data

        delta_fr = fr_int + 4 * np.pi * r * atomic_density

        in_integral = np.array(np.sin(np.outer(q.T, r))) * delta_fr
        integral = np.trapz(in_integral, r) / attenuation_factor
        sq_optimized = sq_int * (1 - 1. / q * integral)

        sq_pattern = Pattern(q, sq_optimized)

        if fcn_callback is not None and iteration % callback_period == 0:
            fr_pattern = calculate_fr(sq_pattern,
                                      use_modification_fcn=use_modification_fcn,
                                      method=fourier_transform_method)
            gr_pattern = calculate_gr_raw(fr_pattern, atomic_density)
            fcn_callback(sq_pattern, fr_pattern, gr_pattern)
    return sq_pattern


def optimize_density(data_pattern, background_pattern, initial_background_scaling, composition,
                     initial_density, background_min, background_max, density_min, density_max,
                     iterations, r_cutoff, use_modification_fcn=False, extrapolation_cutoff=None,
                     r_step=0.01, fcn_callback=None):
    """
    Performs an optimization of the background scaling and density using a figure of merit function defined by the low
    r region in F(r) as described in Eggert et al. (2002) PRB, 65, 174105.

    :param data_pattern:       raw data pattern in Q space (A^-1)
    :param background_pattern: raw background pattern in Q space (A^-1)
    :param initial_background_scaling:
                                start value for the background scaling optimization
    :param composition:         composition of the sample as a dictionary with elements as keys and abundances as values
    :param initial_density:     start value for the density optimization in g/cm^3
    :param background_min:      minimum value for the background scaling
    :param background_max:      maximum value for the background scaling
    :param density_min:         minimum value for the density
    :param density_max:         maximum value for the density
    :param iterations:          number of iterations of S(Q) (see optimize_sq(...) prior to calculating chi2
    :param r_cutoff:            cutoff value below which there is no signal expected (below the first peak in g(r))
    :param use_modification_fcn:
                                Whether to use the Lorch modification function during the Fourier transform.
                                Warning: When using the Lorch modification function, more iterations are needed
                                to get to the wanted result.
    :param extrapolation_cutoff:
                                Determines up to which q value the S(Q) will be extrapolated to zero. The default
                                (None) will use the minimum q value plus 0.2 A^-1
    :param r_step:              Step size for the r-space for calculating f(r) during each iteration.
    :param fcn_callback:        Function which will be called after each iteration. The function should take four
                                arguments: iteration number, chi2, density, and background scaling. Additionally, the
                                function should return a boolean value, where True continues the optimization and False
                                will stop the optimization procedure

    :return: (tuple) - density, density standard error, background scaling, background scaling standard error
    """
    params = lmfit.Parameters()
    params.add("density", value=initial_density, min=density_min, max=density_max)
    params.add("background_scaling", value=initial_background_scaling, min=background_min, max=background_max)

    r = np.arange(0, r_cutoff + r_step / 2., r_step)

    def optimization_fcn(params, extrapolation_max, r, r_cutoff, use_modification_fcn):
        density = params['density'].value
        atomic_density = convert_density_to_atoms_per_cubic_angstrom(composition, density)
        background_pattern.scaling = params['background_scaling'].value

        sq = calculate_sq(data_pattern - background_pattern, density, composition)
        extrapolation_max = extrapolation_max or np.min(sq._x[0]) + 0.2
        sq = extrapolate_to_zero_poly(sq, extrapolation_max)
        sq_optimized = optimize_sq(sq, r_cutoff, iterations, atomic_density, use_modification_fcn)
        fr = calculate_fr(sq_optimized, r=r, use_modification_fcn=use_modification_fcn)

        min_r, min_fr = fr.data

        output = (min_fr + 4 * np.pi * atomic_density * min_r) ** 2 * r_step

        if fcn_callback is not None:
            if not fcn_callback(optimization_fcn.iteration,
                                np.sum(output),
                                density,
                                params['background_scaling'].value):
                return None
        optimization_fcn.iteration += 1
        return output

    optimization_fcn.iteration = 1

    lmfit.minimize(optimization_fcn, params, args=(extrapolation_cutoff, r, r_cutoff, use_modification_fcn))
    lmfit.report_fit(params)

    return params['density'].value, params['density'].stderr, params['background_scaling'].value, \
        params['background_scaling'].stderr


def optimize_incoherent_container_scattering(sample_pattern, sample_density, sample_composition, container_composition,
                                             r_cutoff, initial_content=10, use_extrapolation=True,
                                             extrapolation_q_max=None, callback_fcn=None):
    """
    Finds the amount of extra scattering from a sample container which was not included in the
    background measurement. A typical use-case are diamond anvil cell experiments were the background is usually
    collected for an empty cell with a gasket of a specific thickness. However, during compression, the gasket will
    shrink in thickness and additional diamond compton (incoherent) scattering will be in the resulting data.

    The function tries to achieve this by varying the amount of incoherent scattering from the container and minimizing
    on the intensities of g(r) below a chosen r_cutoff. The r_cutoff parameter should be chosen to be below the first
    peak in g(r) -- usually somewhere between 1 and 1.5 for e.g. silicates and depending on the q_max of the S(Q)

    :param sample_pattern:     Background subtracted data pattern
    :param sample_density:      density of the sample in g/cm^3
    :param sample_composition:  composition of the sample as a dictionary with elements as keys and abundances as values
    :param container_composition:
                                composition of the container_as a dictionary with the elements as keys and the
                                abundances as values
    :param r_cutoff:            an r cutoff for the g(r) in Angstrom for the area used for optimization. Should be
                                below the first peak (basically defines the region where g(r) should be zero for ideal data)
    :param initial_content:     starting content for the optimization
    :param use_extrapolation:   whether to use extrapolation (polynomial) to zero for S(Q) or not prior to transforming it to F(r)
    :param extrapolation_q_max: defines the q range for which the extrapolation to zero will be fitted. Default value
                                (None) which takes the q_min of the sample pattern and adds 0.2 and uses that as a
                                range.
    :param callback_fcn:        function which will be called during each iteration of the optimization. The function s
                                should have an interface for the following parameters:
                                      - background_content - dimensionless number describing the amount of
                                                            incoherent scattering  optimized
                                      - scaled_incoherent_background - calculated scaled incoherent background
                                      - sq - S(Q) calculated using the scaled incoherent background
                                      - fr - F(r) calculated using the scaled incoherent background
                                      - gr - g(r) calculated using the scaled incoherent background

    :return: (tuple) background_content as dimensionless number, scaled incoherent background pattern
    """
    q, _ = sample_pattern.data

    incoherent_background_pattern = Pattern(q, calculate_incoherent_scattering(container_composition, q))
    params = lmfit.Parameters()
    params.add("content", value=initial_content, min=0)

    if extrapolation_q_max is None:
        extrapolation_q_max = np.min(q) + 0.2

    sample_atomic_density = convert_density_to_atoms_per_cubic_angstrom(sample_composition, sample_density)
    sample_incoherent_scattering = calculate_incoherent_scattering(sample_composition, q)
    sample_f_mean_squared = calculate_f_mean_squared(sample_composition, q)
    sample_f_squared_mean = calculate_f_squared_mean(sample_composition, q)

    def optimization_fcn(params):
        background_content = params['content'].value

        incoherent_background_pattern.scaling = background_content
        subtracted_sample_pattern = sample_pattern - incoherent_background_pattern
        sample_normalization_factor = calculate_normalization_factor_raw(
            subtracted_sample_pattern,
            atomic_density=sample_atomic_density,
            f_squared_mean=sample_f_squared_mean,
            f_mean_squared=sample_f_mean_squared,
            incoherent_scattering=sample_incoherent_scattering
        )

        sq = calculate_sq_raw(
            subtracted_sample_pattern,
            f_squared_mean=sample_f_squared_mean,
            f_mean_squared=sample_f_mean_squared,
            incoherent_scattering=sample_incoherent_scattering,
            normalization_factor=sample_normalization_factor
        )

        sq = extrapolate_to_zero_poly(sq, extrapolation_q_max)

        fr = calculate_fr(sq)
        gr = calculate_gr_raw(fr, atomic_density=sample_atomic_density)

        low_r_gr = gr.limit(0, r_cutoff)
        if callback_fcn is not None:
            callback_fcn(background_content, incoherent_background_pattern, sq, fr, gr)

        return low_r_gr.data[1]

    lmfit.minimize(optimization_fcn, params)
    incoherent_background_pattern.scaling = params['content'].value

    return params['content'].value, incoherent_background_pattern


def optimize_soller_dac(data_pattern, bkg_pattern, composition, initial_density, initial_bkg_scaling,
                        initial_thickness, sample_thickness, wavelength,
                        initial_carbon_content=1, r_cutoff=2.28, iterations=1,
                        use_modification_fcn=False, vary=(True, True, True),
                        normalization_method='int', verbose=False):
    """
    Optimizes density, background scaling and diamond content for a list of sample thickness with a given initial
    gasket thickness in the diamond anvil cell (DAC). The calculation is done by utilizing the soller slit transfer
    function and assuming that the DAC has been centered to the rotation center of the soller slit.

    :param data_pattern: original data pattern
    :param bkg_pattern: original background pattern
    :param composition: composition as a dictionary with the elements as keys and the abundances as values
    :param initial_density: number density starting point for the optimization procedure
    :param initial_bkg_scaling: background scaling starting point for the optimization procedure
    :param initial_thickness: gasket thickness with which the background was measured in mm
    :param sample_thickness: sample thickness for which the sample was measured in mm
    :param wavelength: wavelength of the radiation used - needed for calculation of soller slit transfer function in
                       q-space in Angstrom
    :param initial_carbon_content: carbon content starting point for the optimization
    :param r_cutoff: cutoff value below which there is no signal expected (below the first peak in g(r)
    :param iterations: number of iterations for optimization, described in equations 47-49 in Eggert et al. 2002
    :param use_modification_fcn: Whether or not to use the Lorch modification function during the Fourier transform.
    :param vary: 3 boolean flags whether to vary: density, bkg_scaling, carbon_content during the optimization
    :param normalization_method: determines the method used for estimating the normalization method. possible values are
                                 'int' for an integral or 'fit' for fitting the high q region form factors.
    :param verbose: boolean flag whether to print out a fit report or not
    :return:
    """

    q = data_pattern.extend_to(0, 0).x

    f_squared_mean = calculate_f_squared_mean(composition, q)
    f_mean_squared = calculate_f_mean_squared(composition, q)
    incoherent_scattering = calculate_incoherent_scattering(composition, q)

    tth = 2 * np.arcsin(data_pattern.x * wavelength / (4 * np.pi)) / np.pi * 180
    soller = SollerCorrection(tth, initial_thickness)
    sample_transfer, diamond_transfer = soller.transfer_function_dac(sample_thickness, initial_thickness)

    def optimization_fcn(params):
        diamond_content = params['diamond_content'].value
        bkg_scaling = params['bkg_scaling'].value
        density = params['density'].value

        q, data_int = data_pattern.data
        _, bkg_int = bkg_pattern.data

        diamond_background = diamond_content * Pattern(q,
                                                       calculate_incoherent_scattering({'C': 1},
                                                                                       q) / diamond_transfer)

        sample_pattern = data_pattern - bkg_scaling * bkg_pattern
        sample_pattern = sample_pattern - diamond_background
        sample_pattern = Pattern(q, sample_pattern.y * sample_transfer)
        sample_pattern = sample_pattern.extend_to(0, 0)

        if normalization_method == 'fit':
            normalization_factor = fit_normalization_factor(sample_pattern, composition)
        else:
            normalization_factor = calculate_normalization_factor_raw(sample_pattern, density, f_squared_mean,
                                                                      f_mean_squared, incoherent_scattering)

        sq_pattern = calculate_sq_raw(sample_pattern=sample_pattern,
                                      f_squared_mean=f_squared_mean,
                                      f_mean_squared=f_mean_squared,
                                      incoherent_scattering=incoherent_scattering,
                                      normalization_factor=normalization_factor)

        r = np.arange(0, r_cutoff, 0.05)
        fr_pattern = calculate_fr(sq_pattern=sq_pattern, r=r, use_modification_fcn=use_modification_fcn)

        q, sq_int = sq_pattern.data
        r, fr_int = fr_pattern.data
        iq_int = sq_int - 1

        delta_fr = fr_int + 4 * np.pi * r * density

        for iteration in range(iterations):
            in_integral = np.array(np.sin(np.outer(q.T, r))) * delta_fr
            integral = np.trapz(in_integral, r)
            iq_optimized = iq_int - 1. / q * (iq_int + 1) * integral

            iq_pattern = Pattern(q, iq_optimized)
            fr_pattern = calculate_fr(Pattern(q, iq_optimized + 1), r)

            q, iq_int = iq_pattern.data
            r, fr_int = fr_pattern.data

            delta_fr = fr_int + 4 * np.pi * r * density

        return delta_fr / len(delta_fr)

    from lmfit import Parameters, minimize, report_fit

    params = Parameters()
    params.add('density', value=initial_density, min=0, vary=vary[0])
    params.add('bkg_scaling', value=initial_bkg_scaling, vary=vary[1])
    params.add('diamond_content', value=initial_carbon_content, min=0, vary=vary[2])

    result = minimize(optimization_fcn, params)

    if verbose:
        report_fit(result)

    return result.chisqr, \
        result.params['density'].value, result.params['density'].stderr, \
        result.params['bkg_scaling'].value, result.params['bkg_scaling'].stderr, \
        result.params['diamond_content'].value, result.params['diamond_content'].stderr



# -*- coding: utf-8 -*-

import os

from qtpy import QtCore, QtWidgets
import pyqtgraph as pg

from ..widgets.glassure_widget import GlassureWidget
from ..widgets.custom.file_dialogs import open_file_dialog, save_file_dialog

from ..model.glassure_model import GlassureModel
from ..model.configuration import Sample
from ...core.scattering_factors import get_available_elements

from .configuration import ConfigurationController
from .soller import SollerController
from .transfer import TransferFunctionController

# # Switch to using black background and white foreground
pg.setConfigOption('useOpenGL', False)
pg.setConfigOption('leftButtonPan', False)
pg.setConfigOption('background', 'k')
pg.setConfigOption('foreground', 'w')
pg.setConfigOption('antialias', True)


class GlassureController(object):
    def __init__(self):
        self.main_widget = GlassureWidget()

        self.model = GlassureModel()
        self.settings = QtCore.QSettings('Glassure', 'Glassure')
        self.connect_signals()

        self.configuration_controller = ConfigurationController(self.main_widget, self.model)
        self.soller_controller = SollerController(self.main_widget, self.model)
        self.transfer_controller = TransferFunctionController(self.main_widget, self.model)

    def show_window(self):
        """
        Displays the main window on the screen and makes it active
        """
        self.main_widget.show()

    def connect_signals(self):
        """
        Connects Gui signals with the model and model signals with the GUI.
        """

        # model
        self.model.data_changed.connect(self.model_changed)

        self.main_widget.load_data_btn.clicked.connect(self.load_data)
        self.main_widget.load_bkg_btn.clicked.connect(self.load_bkg)
        self.main_widget.reset_bkg_btn.clicked.connect(self.reset_bkg)

        # connecting background scaling and smoothing of the original data
        self.main_widget.bkg_scaling_sb.valueChanged.connect(self.bkg_scale_changed)
        self.main_widget.smooth_sb.valueChanged.connect(self.smooth_changed)

        # updating the composition
        self.main_widget.left_control_widget.composition_widget.composition_changed.connect(self.update_sample)
        self.main_widget.add_element_btn.clicked.connect(self.add_element_btn_clicked)
        self.main_widget.delete_element_btn.clicked.connect(self.delete_element_btn_clicked)
        self.main_widget.left_control_widget.composition_widget.composition_changed.connect(self.update_model)

        # updating the calculation parameters
        self.main_widget.left_control_widget.options_widget.options_parameters_changed.connect(self.update_model)
        self.main_widget.left_control_widget.extrapolation_widget.extrapolation_parameters_changed.connect(
            self.update_model)

        # optimization controls
        optimization_widget = self.main_widget.right_control_widget.optimization_widget
        optimization_widget.plot_progress_cb.stateChanged.connect(self.update_plot_progress)
        optimization_widget.optimization_parameters_changed.connect(self.update_model)

        self.main_widget.right_control_widget.density_optimization_widget.optimize_btn.clicked.connect(
            self.optimize_density)

        # Diamond controls
        self.main_widget.right_control_widget.diamond_widget.diamond_txt.editingFinished.connect(
            self.diamond_content_changed)
        self.main_widget.right_control_widget.diamond_widget.diamond_optimize_btn.clicked.connect(
            self.optimize_diamond_btn_clicked
        )

        # Saving the resulting data
        self.main_widget.save_sq_btn.clicked.connect(self.save_sq_btn_clicked)
        self.main_widget.save_gr_btn.clicked.connect(self.save_gr_btn_clicked)

    def load_data(self):
        filename = open_file_dialog(self.main_widget, caption="Load Pattern",
                                    directory=self.settings.value('working_directory'))

        if filename != '':
            self.model.load_data(filename)
            self.settings.setValue('working_directory', os.path.dirname(filename))
            self.main_widget.left_control_widget.data_widget.file_widget.data_filename_lbl.setText(
                self.model.current_configuration.original_pattern.name)

    def load_bkg(self):
        filename = open_file_dialog(self.main_widget, "Load background data",
                                    directory=self.settings.value('working_directory'))

        if filename is not None and filename != '':
            self.model.load_bkg(filename)
            self.settings.setValue('working_directory', os.path.dirname(filename))
            self.main_widget.left_control_widget.data_widget.file_widget.background_filename_lbl.setText(
                self.model.current_configuration.background_pattern.name)

    def reset_bkg(self):
        self.model.reset_bkg()

    def model_changed(self):
        if self.model.original_pattern is not None:
            self.main_widget.pattern_widget.plot_pattern(self.model.original_pattern)
        self.main_widget.pattern_widget.plot_bkg(self.model.background_pattern)
        if self.model.sq_pattern is not None:
            self.main_widget.pattern_widget.set_sq_pattern(self.model.sq_pattern, self.model.configuration_ind)
        if self.model.gr_pattern is not None:
            self.main_widget.pattern_widget.set_gr_pattern(self.model.gr_pattern, self.model.configuration_ind)

        if self.model.background_pattern is not None:
            self.main_widget.bkg_filename_lbl.setText(self.model.background_pattern.name)
            self.main_widget.bkg_scaling_sb.setEnabled(True)
            self.main_widget.bkg_scaling_sb.setValue(self.model.background_pattern.scaling)
        else:
            self.main_widget.bkg_filename_lbl.setText('None')
            self.main_widget.bkg_scaling_sb.setEnabled(False)
        self.main_widget.smooth_sb.setValue(self.model.original_pattern.smoothing)

        self.main_widget.left_control_widget.composition_widget.density_atomic_units_lbl. \
            setText("{:.4f}".format(self.model.atomic_density))

        self.main_widget.update_sample_config(self.model.sample)
        self.main_widget.update_transform_config(self.model.transform_config)
        self.main_widget.update_extrapolation_config(self.model.extrapolation_config)

    def bkg_scale_changed(self, value):
        self.model.background_scaling = value

    def update_bkg_scale_step(self):
        value = float(self.main_widget.bkg_scale_step_txt.text())
        self.main_widget.bkg_scale_sb.setSingleStep(value)

    def update_bkg_offset_step(self):
        value = float(self.main_widget.bkg_offset_step_txt.text())
        self.main_widget.bkg_offset_sb.setSingleStep(value)

    def smooth_changed(self, value):
        self.model.set_smooth(value)

    def update_smooth_step(self):
        value = float(self.main_widget.smooth_step_txt.text())
        self.main_widget.smooth_sb.setSingleStep(value)

    def add_element_btn_clicked(self):
        self.model.add_element()

    def delete_element_btn_clicked(self):
        cur_ind = self.main_widget.left_control_widget.composition_widget.composition_tw.currentRow()
        self.main_widget.left_control_widget.composition_widget.delete_element(cur_ind)

    def update_sample(self, sample: Sample):
        self.model.sample_config = sample

    def update_model(self):
        sample_config = self.main_widget.get_sample_config()
        transform_config = self.main_widget.get_transform_config()
        extrapolation_config = self.main_widget.get_extrapolation_config()

        optimize_active, r_cutoff, optimize_iterations, optimize_attenuation = \
            self.main_widget.get_optimization_parameter()

        self.model.update_parameter(sample_config,
                                    transform_config,
                                    extrapolation_config,
                                    optimize_active,
                                    r_cutoff,
                                    optimize_iterations,
                                    optimize_attenuation
                                    )

    def update_plot_progress(self, bool):
        if bool:
            self.model.optimization_callback = self.plot_optimization_progress
        else:
            self.model.optimization_callback = None

    def plot_optimization_progress(self, sq_pattern, fr_pattern, gr_pattern):
        self.main_widget.pattern_widget.set_sq_pattern(sq_pattern, self.model.configuration_ind)
        self.main_widget.pattern_widget.set_gr_pattern(gr_pattern, self.model.configuration_ind)
        QtWidgets.QApplication.processEvents()

    def optimize_density(self):
        density_min, density_max, bkg_min, bkg_max, iterations = \
            self.main_widget.right_control_widget.density_optimization_widget.get_parameters()
        self.model.optimize_density_and_scaling(
            density_min, density_max, bkg_min, bkg_max, iterations, output_txt=
            self.main_widget.right_control_widget.density_optimization_widget.optimization_output_txt,
            callback_fcn=self.plot_optimization_progress)

    def diamond_content_changed(self):
        new_value = float(str(self.main_widget.right_control_widget.diamond_widget.diamond_txt.text()))
        self.model.set_diamond_content(new_value)

    def optimize_diamond_btn_clicked(self):
        start_value = float(str(self.main_widget.right_control_widget.diamond_widget.diamond_txt.text()))

        def callback_fcn(diamond_content):
            self.main_widget.right_control_widget.diamond_widget.diamond_txt.setText('{:.2f}'.format(diamond_content))
            QtWidgets.QApplication.processEvents()

        self.model.optimize_diamond_content(diamond_content=start_value, callback_fcn=callback_fcn)

    def save_sq_btn_clicked(self):
        if self.settings.value('sq_directory') is not None:
            sq_filename = os.path.join(self.settings.value('sq_directory'),
                                       self.model.original_pattern.name + ".txt")
        else:
            sq_filename = None
        filename = save_file_dialog(self.main_widget,
                                    "Save S(Q) Data.",
                                    sq_filename,
                                    ('Data (*.txt)'))
        if filename != '':
            self.model.sq_pattern.save(filename)
            self.settings.setValue('sq_directory', os.path.dirname(filename))

    def save_gr_btn_clicked(self):
        if self.settings.value('gr_directory') is not None:
            gr_filename = os.path.join(self.settings.value('gr_directory'),
                                       self.model.original_pattern.name + ".txt")
        else:
            gr_filename = None
        filename = save_file_dialog(self.main_widget,
                                    "Save g(r) Data.",
                                    gr_filename,
                                    ('Data (*.txt)'))
        if filename != '':
            self.model.gr_pattern.save(filename)
            self.settings.setValue('gr_directory', os.path.dirname(filename))



# -*- coding: utf-8 -*-

from qtpy import QtCore, QtWidgets
from ..custom import HorizontalLine, FloatLineEdit, DragSlider
import numpy as np
from ...model.configuration import TransformConfiguration
from ....core.methods import SqMethod, NormalizationMethod, FourierTransformMethod


class OptionsWidget(QtWidgets.QWidget):
    options_parameters_changed = QtCore.Signal()

    def __init__(self, *args):
        super(OptionsWidget, self).__init__(*args)

        self.create_widgets()
        self.style_widgets()
        self.create_layout()
        self.create_signals()

    def create_widgets(self):
        self.calculation_ranges_gb = QtWidgets.QGroupBox("Ranges")
        self.q_range_lbl = QtWidgets.QLabel('Q:')
        self.q_min_txt = FloatLineEdit('0')
        self.q_max_txt = FloatLineEdit('10')
        self.q_max_slider = DragSlider()

        self.r_range_lbl = QtWidgets.QLabel('r:')
        self.r_min_txt = FloatLineEdit('0.5')
        self.r_max_txt = FloatLineEdit('10')

        self.modification_fcn_cb = QtWidgets.QCheckBox("Use Modification Function")
        self.fft_cb = QtWidgets.QCheckBox("Use FFT")
        self.fft_cb.setToolTip(
            "Use FFT for Fourier Transform. If not checked, the Fourier integral is solver numerically.")

        self.normalization_method_gb = QtWidgets.QGroupBox("Normalization")
        self.normalization_method_integral = QtWidgets.QRadioButton("Integral")
        self.normalization_method_integral.setChecked(True)
        self.normalization_method_fit = QtWidgets.QRadioButton("Fit")
        self.normalization_method_fit.setChecked(False)

        self.sq_method_gb = QtWidgets.QGroupBox("S(Q) Method")
        self.sq_method_FZ = QtWidgets.QRadioButton("Faber-Ziman")
        self.sq_method_FZ.setChecked(True)
        self.sq_method_AL = QtWidgets.QRadioButton("Ashcroft-Langreth")
        self.sq_method_AL.setChecked(False)

    def style_widgets(self):
        self.q_range_lbl.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.r_range_lbl.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)

        self.sq_method_FZ.setLayoutDirection(QtCore.Qt.RightToLeft)
        self.sq_method_AL.setLayoutDirection(QtCore.Qt.RightToLeft)
        self.normalization_method_integral.setLayoutDirection(QtCore.Qt.RightToLeft)
        self.normalization_method_fit.setLayoutDirection(QtCore.Qt.RightToLeft)

        self.modification_fcn_cb.setLayoutDirection(QtCore.Qt.RightToLeft)
        self.fft_cb.setLayoutDirection(QtCore.Qt.RightToLeft)

    def create_layout(self):
        self.main_layout = QtWidgets.QVBoxLayout()
        self.main_layout.setSpacing(5)
        self.main_layout.setContentsMargins(0, 0, 0, 0)

        self._calculation_ranges_layout = QtWidgets.QGridLayout()
        self._calculation_ranges_layout.setSpacing(5)

        self._calculation_ranges_layout.addWidget(self.q_range_lbl, 0, 1)
        self._calculation_ranges_layout.addWidget(self.q_min_txt, 0, 2)
        self._calculation_ranges_layout.addWidget(QtWidgets.QLabel('-'), 0, 3)
        self._calculation_ranges_layout.addWidget(self.q_max_txt, 0, 4)
        self._calculation_ranges_layout.addWidget(QtWidgets.QLabel('A<sup>-1</sup>'), 0, 5)

        self._calculation_ranges_layout.addWidget(self.q_max_slider, 1, 4, 1, 1)

        self._calculation_ranges_layout.addWidget(self.r_range_lbl, 2, 1)
        self._calculation_ranges_layout.addWidget(self.r_min_txt, 2, 2)
        self._calculation_ranges_layout.addWidget(QtWidgets.QLabel('-'), 2, 3)
        self._calculation_ranges_layout.addWidget(self.r_max_txt, 2, 4)
        self._calculation_ranges_layout.addWidget(QtWidgets.QLabel('A'), 2, 5)
        self.calculation_ranges_gb.setLayout(self._calculation_ranges_layout)
        self.main_layout.addWidget(self.calculation_ranges_gb)

        self.choice_layout = QtWidgets.QHBoxLayout()
        self.choice_layout.setSpacing(5)

        self._normalization_method_gb_layout = QtWidgets.QVBoxLayout()
        self._normalization_method_gb_layout.setSpacing(5)
        self._normalization_method_gb_layout.addWidget(self.normalization_method_integral)
        self._normalization_method_gb_layout.addWidget(self.normalization_method_fit)
        self.normalization_method_gb.setLayout(self._normalization_method_gb_layout)
        self.choice_layout.addWidget(self.normalization_method_gb)

        self._sq_method_gb_layout = QtWidgets.QVBoxLayout()
        self._sq_method_gb_layout.setSpacing(5)
        self._sq_method_gb_layout.addWidget(self.sq_method_FZ)
        self._sq_method_gb_layout.addWidget(self.sq_method_AL)
        self.sq_method_gb.setLayout(self._sq_method_gb_layout)
        self.choice_layout.addWidget(self.sq_method_gb)

        self.main_layout.addLayout(self.choice_layout)
        self.main_layout.addWidget(self.modification_fcn_cb)
        self.main_layout.addWidget(self.fft_cb)

        self.setLayout(self.main_layout)

    def create_signals(self):
        self.q_max_txt.editingFinished.connect(self.txt_changed)
        self.q_min_txt.editingFinished.connect(self.txt_changed)
        self.q_max_slider.dragChanged.connect(self.q_max_slider_changed)
        self.r_min_txt.editingFinished.connect(self.txt_changed)
        self.r_max_txt.editingFinished.connect(self.txt_changed)

        self.modification_fcn_cb.stateChanged.connect(self.options_changed)
        self.fft_cb.stateChanged.connect(self.options_changed)
        self.sq_method_FZ.toggled.connect(self.options_changed)
        self.normalization_method_integral.toggled.connect(self.options_changed)

    def txt_changed(self):
        if self.q_max_txt.isModified() or self.q_min_txt.isModified() or \
                self.r_min_txt.isModified() or self.r_max_txt.isModified():
            self.options_parameters_changed.emit()

            self.q_max_txt.setModified(False)
            self.q_min_txt.setModified(False)
            self.r_min_txt.setModified(False)
            self.r_max_txt.setModified(False)

    def q_max_slider_changed(self):
        q_max = float(self.q_max_txt.text())
        factor = np.tan(self.q_max_slider.value() / 75) / 150 + 1
        self.q_max_txt.setText(f"{factor * q_max:.2f}")
        self.options_changed()

    def options_changed(self):
        self.options_parameters_changed.emit()

    def get_ranges(self):
        q_min = self.q_min_txt.value()
        q_max = self.q_max_txt.value()
        r_min = self.r_min_txt.value()
        r_max = self.r_max_txt.value()
        return q_min, q_max, r_min, r_max

    def get_normalization_method(self):
        if self.normalization_method_integral.isChecked():
            return NormalizationMethod.INTEGRAL
        elif self.normalization_method_fit.isChecked():
            return NormalizationMethod.FIT
        else:
            return None

    def get_sq_method(self):
        if self.sq_method_FZ.isChecked():
            return SqMethod.FZ
        elif self.sq_method_AL.isChecked():
            return SqMethod.AL
        else:
            return None

    def get_fourier_transform_method(self):
        if self.fft_cb.isChecked():
            return FourierTransformMethod.FFT
        else:
            return FourierTransformMethod.INTEGRAL

    def set_fourier_transform_method(self, method):
        if method == 'fft' or method == FourierTransformMethod.FFT:
            self.fft_cb.setChecked(True)
        else:
            self.fft_cb.setChecked(False)

    def get_transform_configuration(self) -> TransformConfiguration:
        config = TransformConfiguration()
        config.q_min, config.q_max, config.r_min, config.r_max = self.get_ranges()
        config.normalization_method = self.get_normalization_method()
        config.sq_method = self.get_sq_method()
        config.use_modification_fcn = self.modification_fcn_cb.isChecked()
        config.fourier_transform_method = self.get_fourier_transform_method()

        return config

    def update_transform_configuration(self, config: TransformConfiguration):
        self.blockSignals(True)
        self.q_min_txt.setText(f"{config.q_min:.2f}")
        self.q_max_txt.setText(f"{config.q_max:.2f}")
        self.r_min_txt.setText(f"{config.r_min:.2f}")
        self.r_max_txt.setText(f"{config.r_max:.2f}")

        self.modification_fcn_cb.blockSignals(True)
        self.modification_fcn_cb.setChecked(config.use_modification_fcn)
        self.modification_fcn_cb.blockSignals(False)

        if config.normalization_method == NormalizationMethod.INTEGRAL:
            self.normalization_method_integral.setChecked(True)
        elif config.normalization_method == NormalizationMethod.FIT:
            self.normalization_method_fit.setChecked(True)
        if config.sq_method == SqMethod.FZ:
            self.sq_method_FZ.setChecked(True)
        elif config.sq_method == SqMethod.AL:
            self.sq_method_AL.setChecked(True)

        self.set_fourier_transform_method(config.fourier_transform_method)
        self.blockSignals(False)



# -*- coding: utf-8 -*-

from qtpy import QtCore, QtGui, QtWidgets

from ..custom import HorizontalLine, HorizontalSpacerItem, FloatLineEdit, LabelAlignRight
from ...model.configuration import ExtrapolationConfiguration


class ExtrapolationWidget(QtWidgets.QWidget):
    extrapolation_parameters_changed = QtCore.Signal(ExtrapolationConfiguration)

    def __init__(self, *args):
        super(ExtrapolationWidget, self).__init__(*args)

        self.create_widgets()
        self.create_layout()
        self.style_widgets()
        self.create_signals()

        self.update_visibility()
        self.rb_widget.setVisible(False)
        self.activate_cb.setChecked(True)

    def create_widgets(self):
        self.activate_cb = QtWidgets.QCheckBox("activate")

        self.step_extrapolation_rb = MyRadioButton('Step')
        self.linear_extrapolation_rb = MyRadioButton("Linear")
        self.poly_extrapolation_rb = MyRadioButton("Polynomial")
        self.spline_extrapolation_rb = MyRadioButton("Spline")
        self.step_extrapolation_rb.setChecked(True)

        self.s0_label = LabelAlignRight("S(Q=0):")
        self.s0_txt = FloatLineEdit("0")
        self.s0_auto_cb = QtWidgets.QCheckBox("auto")
        self.s0_auto_cb.setToolTip("Automatically determines S(Q=0) from the form factors.")
        self.s0_auto_cb.setChecked(True)

        self.options_gb = QtWidgets.QGroupBox(self)
        self.q_max_lbl = LabelAlignRight("Fit Q_max:")
        self.q_max_txt = FloatLineEdit("2")
        self.q_max_unit_lbl = LabelAlignRight("A")
        self.replace_cb = QtWidgets.QCheckBox("replace")
        self.replace_cb.setToolTip("Replace data with extrapolated data in overlapping region.")

        self.rb_extrapolation_button_group = QtWidgets.QButtonGroup()
        self.rb_extrapolation_button_group.addButton(self.step_extrapolation_rb)
        self.rb_extrapolation_button_group.addButton(self.linear_extrapolation_rb)
        self.rb_extrapolation_button_group.addButton(self.poly_extrapolation_rb)
        self.rb_extrapolation_button_group.addButton(self.spline_extrapolation_rb)

    def style_widgets(self):
        self.q_max_lbl.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.q_max_txt.setMaximumWidth(50)
        self.s0_txt.setMaximumWidth(50)

    def create_layout(self):
        self.vertical_layout = QtWidgets.QVBoxLayout()
        self.vertical_layout.setContentsMargins(0, 0, 0, 0)
        self.vertical_layout.setSpacing(5)

        self.vertical_layout.addWidget(self.activate_cb)
        self.vertical_layout.addWidget(HorizontalLine())

        self.rb_layout = QtWidgets.QGridLayout()
        self.rb_layout.setContentsMargins(5, 5, 5, 5)
        self.rb_layout.setSpacing(8)

        self.rb_layout.addWidget(self.step_extrapolation_rb, 0, 0)
        self.rb_layout.addWidget(self.linear_extrapolation_rb, 0, 1)
        self.rb_layout.addWidget(self.poly_extrapolation_rb, 1, 0)
        self.rb_layout.addWidget(self.spline_extrapolation_rb, 1, 1)

        self.rb_widget = QtWidgets.QWidget(self)
        self.rb_widget.setLayout(self.rb_layout)
        self.vertical_layout.addWidget(self.rb_widget)

        self.options_layout = QtWidgets.QGridLayout()
        self.options_layout.setSpacing(5)
        self.options_layout.setColumnStretch(0, 1)

        self.options_layout.addWidget(self.s0_label, 0, 1)
        self.options_layout.addWidget(self.s0_txt, 0, 2)
        self.options_layout.addWidget(self.s0_auto_cb, 0, 4)

        self.options_layout.addWidget(self.q_max_lbl, 1, 1)
        self.options_layout.addWidget(self.q_max_txt, 1, 2)
        self.options_layout.addWidget(self.q_max_unit_lbl, 1, 3)
        self.options_layout.addWidget(self.replace_cb, 1, 4)

        self.options_gb.setLayout(self.options_layout)

        self.vertical_layout.addWidget(self.options_gb)
        self.setLayout(self.vertical_layout)

    def create_signals(self):
        self.activate_cb.stateChanged.connect(self.rb_widget.setVisible)
        self.activate_cb.stateChanged.connect(self.options_gb.setVisible)
        self.activate_cb.stateChanged.connect(self.emit_changed_signal)

        self.rb_extrapolation_button_group.buttonReleased.connect(self.emit_changed_signal)
        self.rb_extrapolation_button_group.buttonReleased.connect(self.update_visibility)

        self.q_max_txt.editingFinished.connect(self.emit_changed_signal)
        self.replace_cb.stateChanged.connect(self.emit_changed_signal)

        self.s0_txt.editingFinished.connect(self.emit_changed_signal)
        self.s0_auto_cb.stateChanged.connect(self.emit_changed_signal)

    def emit_changed_signal(self):
        self.extrapolation_parameters_changed.emit(self.get_configuration())

    def update_visibility(self):
        visible_flag = not (self.step_extrapolation_rb.isChecked() | self.linear_extrapolation_rb.isChecked())
        self.q_max_lbl.setVisible(visible_flag)
        self.q_max_unit_lbl.setVisible(visible_flag)
        self.q_max_txt.setVisible(visible_flag)
        self.replace_cb.setVisible(visible_flag)

    def get_configuration(self) -> ExtrapolationConfiguration:
        config = ExtrapolationConfiguration()
        config.activate = self.activate_cb.isChecked()
        config.method = self.get_method()
        config.fit_q_max = float(str(self.q_max_txt.text()))
        config.fit_replace = self.replace_cb.isChecked()
        config.s0 = float(str(self.s0_txt.text()))
        config.s0_auto = self.s0_auto_cb.isChecked()
        return config

    def update_configuration(self, config: ExtrapolationConfiguration):
        self.set_extrapolation_method(config.method)
        self.q_max_txt.setText(f"{config.fit_q_max:.2f}")

        self.activate_cb.blockSignals(True)
        self.s0_auto_cb.blockSignals(True)
        self.replace_cb.blockSignals(True) 

        self.activate_cb.setChecked(config.activate)
        self.replace_cb.setChecked(config.fit_replace)
        self.s0_txt.setText(f"{config.s0:.2f}")
        self.s0_auto_cb.setChecked(config.s0_auto)

        self.activate_cb.blockSignals(False)
        self.replace_cb.blockSignals(False) 
        self.s0_auto_cb.blockSignals(False)
        self.update_visibility()

    def get_method(self):
        if self.step_extrapolation_rb.isChecked():
            return 'step'
        elif self.linear_extrapolation_rb.isChecked():
            return "linear"
        elif self.poly_extrapolation_rb.isChecked():
            return "poly"
        elif self.spline_extrapolation_rb.isChecked():
            return "spline"

    def set_extrapolation_method(self, method):
        self.rb_extrapolation_button_group.blockSignals(True)
        if method == 'step':
            self.step_extrapolation_rb.setChecked(True)
        elif method == "linear":
            self.linear_extrapolation_rb.setChecked(True)
        elif method == "poly":
            self.poly_extrapolation_rb.setChecked(True)
        elif method == "spline":
            self.spline_extrapolation_rb.setChecked(True)
        self.rb_extrapolation_button_group.blockSignals(False)

        self.update_visibility()


class MyRadioButton(QtWidgets.QPushButton):
    def __init__(self, *args):
        super(MyRadioButton, self).__init__(*args)
        self.setCheckable(True)
        self.setFlat(True)
        self.setMinimumHeight(25)



# -*- coding: utf-8 -*-
from qtpy import QtCore, QtWidgets

from ..custom import FloatLineEdit, CommaDoubleValidator
from ....core.scattering_factors import calculators, sources
from ...model.configuration import Sample

Signal = QtCore.Signal


class CompositionWidget(QtWidgets.QWidget):
    composition_changed = Signal(Sample)

    def __init__(self, *args):
        super(CompositionWidget, self).__init__(*args)

        self._create_widgets()
        self._create_layout()
        self._connect_signals()
        self._style_widgets()

    def _create_widgets(self):
        self.source_lbl = QtWidgets.QLabel("Source:")
        self.source_cb = QtWidgets.QComboBox()
        self.source_cb.addItems(sources)
        self.source_cb.setCurrentIndex(0)

        self.add_element_btn = QtWidgets.QPushButton("Add")
        self.delete_element_btn = QtWidgets.QPushButton("Delete")

        self.density_lbl = QtWidgets.QLabel("Density:")
        self.density_txt = FloatLineEdit("2.2")
        self.density_atomic_units_lbl = QtWidgets.QLabel("")

        self.composition_tw = QtWidgets.QTableWidget()

    def _create_layout(self):
        self.main_layout = QtWidgets.QVBoxLayout()
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(5)

        self.source_layout = QtWidgets.QHBoxLayout()
        self.source_layout.setSpacing(5)
        self.source_layout.addWidget(self.source_lbl)
        self.source_layout.addWidget(self.source_cb)

        self.button_layout = QtWidgets.QHBoxLayout()
        self.button_layout.setSpacing(15)
        self.button_layout.addWidget(self.add_element_btn)
        self.button_layout.addWidget(self.delete_element_btn)

        self.density_layout = QtWidgets.QGridLayout()
        self.density_layout.addWidget(self.density_lbl, 0, 0)
        self.density_layout.addWidget(self.density_txt, 0, 1)
        self.density_layout.addWidget(QtWidgets.QLabel('g/cm^3'), 0, 2)
        self.density_layout.addWidget(self.density_atomic_units_lbl, 1, 1)
        self.density_layout.addWidget(QtWidgets.QLabel('at/A^3'), 1, 2)

        self.main_layout.addLayout(self.source_layout)
        self.main_layout.addLayout(self.button_layout)
        self.main_layout.addWidget(self.composition_tw)
        self.main_layout.addLayout(self.density_layout)
        self.setLayout(self.main_layout)

    def _connect_signals(self):
        self.source_cb.currentTextChanged.connect(self.emit_composition_changed_signal)
        self.density_txt.editingFinished.connect(self.emit_composition_changed_signal)

        self.composition_tw.cellChanged.connect(self.emit_composition_changed_signal)

    def _style_widgets(self):
        self.source_lbl.setAlignment(QtCore.Qt.AlignVCenter | QtCore.Qt.AlignRight)
        self.density_lbl.setAlignment(QtCore.Qt.AlignVCenter | QtCore.Qt.AlignRight)
        self.density_atomic_units_lbl.setAlignment(QtCore.Qt.AlignVCenter | QtCore.Qt.AlignRight)

        self.density_txt.editingFinished.connect(self.emit_composition_changed_signal)

        self.density_txt.setMaximumWidth(100)

        self.composition_tw.setColumnCount(2)
        self.composition_tw.horizontalHeader().setVisible(False)
        self.composition_tw.verticalHeader().setVisible(False)
        self.composition_tw.setColumnWidth(0, 80)
        self.composition_tw.setColumnWidth(1, 80)
        self.composition_tw.setItemDelegate(TextDoubleDelegate(self))

    def add_element(self, element=None, value=None):
        current_rows = self.composition_tw.rowCount()
        self.composition_tw.blockSignals(True)
        self.composition_tw.setRowCount(current_rows + 1)

        element_cb = QtWidgets.QComboBox(self)
        element_cb.setStyle(QtWidgets.QStyleFactory.create('cleanlooks'))

        for ind, ele in enumerate(calculators[self.source_cb.currentText()].elements):
            element_cb.insertItem(ind, ele)

        if element is not None:
            element_cb.setCurrentIndex(element_cb.findText(element))

        self.composition_tw.setCellWidget(current_rows, 0, element_cb)

        element_cb.currentIndexChanged.connect(self.emit_composition_changed_signal)

        if value is not None:
            value_item = QtWidgets.QTableWidgetItem(str(value))
        else:
            value_item = QtWidgets.QTableWidgetItem(str(1))

        value_item.setTextAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)

        self.composition_tw.setItem(current_rows, 1, value_item)
        self.composition_tw.blockSignals(False)
        self.emit_composition_changed_signal()

    def delete_element(self, ind):
        self.composition_tw.blockSignals(True)
        # If no row is selected, delete the last row
        if ind == -1:
            ind = self.composition_tw.rowCount() - 1
        self.composition_tw.removeRow(ind)

        # Select the next row
        if ind < self.composition_tw.rowCount():
            self.composition_tw.selectRow(ind)
        else:
            self.composition_tw.selectRow(ind - 1)
        self.composition_tw.blockSignals(False)
        self.emit_composition_changed_signal()

    def set_composition(self, composition):
        self.composition_tw.blockSignals(True)
        self.blockSignals(True)
        self.composition_tw.setRowCount(0)
        for element, value in composition.items():
            self.add_element(element, value)
        self.composition_tw.blockSignals(False)
        self.blockSignals(False)

    def get_composition(self):
        composition = {}

        for row_ind in range(self.composition_tw.rowCount()):
            cb_item = self.composition_tw.cellWidget(row_ind, 0)
            value_item = self.composition_tw.item(row_ind, 1)
            composition[str(cb_item.currentText())] = float(str(value_item.text().replace(',', '.')))

        return composition

    def get_sample_configuration(self) -> Sample:
        sample = Sample()
        sample.composition = self.get_composition()
        sample.density = self.density_txt.value()
        sample.sf_source = str(self.source_cb.currentText())
        return sample

    def set_sf_source(self, source):
        self.source_cb.blockSignals(True)
        self.source_cb.setCurrentIndex(self.source_cb.findText(source))
        self.source_cb.blockSignals(False)

    def update_sample_configuration(self, sample: Sample):
        self.set_sf_source(sample.sf_source)
        self.set_composition(sample.composition)
        self.density_txt.setText(f"{sample.density:.2f}")

    def emit_composition_changed_signal(self):
        self.composition_changed.emit(self.get_sample_configuration())


class TextDoubleDelegate(QtWidgets.QStyledItemDelegate):
    def __init__(self, parent):
        super(TextDoubleDelegate, self).__init__(parent)

    def createEditor(self, parent, _, model):
        self.editor = QtWidgets.QLineEdit(parent)
        self.editor.setFrame(False)
        self.editor.setValidator(CommaDoubleValidator())
        self.editor.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        return self.editor

    def setEditorData(self, parent, index):
        value = index.model().data(index, QtCore.Qt.EditRole)
        try:
            value = value.toString()
        except AttributeError:
            value = value

        if value != '':
            self.editor.setText("{:g}".format(float(value)))

    def setModelData(self, parent, model, index):
        value = self.editor.text()
        model.setData(index, value, QtCore.Qt.EditRole)

    def updateEditorGeometry(self, editor, option, _):
        editor.setGeometry(option.rect)



# -*- coding: utf-8 -*-
from qtpy import QtCore, QtWidgets
from ..custom import HorizontalLine, FloatLineEdit, IntegerLineEdit


class DensityOptimizationWidget(QtWidgets.QWidget):
    calculation_parameters_changed = QtCore.Signal(float)

    def __init__(self, *args):
        super(DensityOptimizationWidget, self).__init__(*args)
        self.create_widgets()
        self.style_widgets()
        self.create_layout()

    def create_widgets(self):
        self.density_range_lbl = QtWidgets.QLabel('Density range:')
        self.density_min_txt = FloatLineEdit('1')
        self.density_max_txt = FloatLineEdit('10')

        self.bkg_range_lbl = QtWidgets.QLabel("Bkg range:")
        self.bkg_min_txt = FloatLineEdit('0.1')
        self.bkg_max_txt = FloatLineEdit('2')

        self.optimize_btn = QtWidgets.QPushButton("Optimize")
        self.optimize_iterations_lbl = QtWidgets.QLabel("Iterations:")

        self.optimize_iterations_txt = IntegerLineEdit("5")
        self.optimization_output_txt = QtWidgets.QTextEdit()
        self.optimization_output_txt.setReadOnly(True)

    def style_widgets(self):
        center_right = QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter
        self.density_range_lbl.setAlignment(center_right)
        self.bkg_range_lbl.setAlignment(center_right)
        self.optimize_iterations_lbl.setAlignment(center_right)

        self.optimize_iterations_txt.setMaximumWidth(80)

        self.optimize_btn.setFlat(True)

    def create_layout(self):
        self.grid_layout = QtWidgets.QGridLayout()
        self.grid_layout.setContentsMargins(0, 0, 0, 0)
        self.grid_layout.setSpacing(5)

        self.grid_layout.addWidget(self.density_range_lbl, 0, 0)
        self.grid_layout.addWidget(self.density_min_txt, 0, 1)
        self.grid_layout.addWidget(QtWidgets.QLabel('-'), 0, 2)
        self.grid_layout.addWidget(self.density_max_txt, 0, 3)
        self.grid_layout.addWidget(QtWidgets.QLabel("g/cm^3"), 0, 4)

        self.grid_layout.addWidget(self.bkg_range_lbl, 1, 0)
        self.grid_layout.addWidget(self.bkg_min_txt, 1, 1)
        self.grid_layout.addWidget(QtWidgets.QLabel('-'), 1, 2)
        self.grid_layout.addWidget(self.bkg_max_txt, 1, 3)

        self.grid_layout.addWidget(self.optimize_iterations_lbl, 2, 0)
        self.grid_layout.addWidget(self.optimize_iterations_txt, 2, 1)

        self.grid_layout.addWidget(self.optimize_btn, 3, 0, 1, 5)
        self.grid_layout.addWidget(self.optimization_output_txt, 4, 0, 1, 5)

        self.setLayout(self.grid_layout)

    def get_parameters(self):
        density_min = self.density_min_txt.value()
        density_max = self.density_max_txt.value()
        bkg_min = self.bkg_min_txt.value()
        bkg_max = self.bkg_max_txt.value()
        iterations = self.optimize_iterations_txt.value()
        return density_min, density_max, bkg_min, bkg_max, iterations



# -*- coding: utf-8 -*-
from qtpy import QtCore, QtWidgets

from ..custom import FloatLineEdit, IntegerLineEdit, LabelAlignRight
from ..custom.lines import HorizontalLine

Signal = QtCore.Signal


class OptimizationWidget(QtWidgets.QWidget):
    optimization_parameters_changed = Signal()

    def __init__(self, *args):
        super(OptimizationWidget, self).__init__(*args)

        self.create_widgets()
        self.style_widgets()
        self.create_layout()
        self.create_signals()

        self.param_widget.setVisible(False)
        self.activate_cb.setChecked(False)

    def create_widgets(self):
        self.activate_cb = QtWidgets.QCheckBox("activate")

        self.r_cutoff_lbl = LabelAlignRight('r cutoff:')
        self.r_cutoff_txt = FloatLineEdit('1.4')

        self.optimize_iterations_lbl = LabelAlignRight("Iterations:")
        self.optimize_iterations_txt = IntegerLineEdit('5')

        self.attenuation_factor_lbl = LabelAlignRight("Attenuation:")
        self.attenuation_factor_sb = QtWidgets.QSpinBox()

        self.plot_progress_cb = QtWidgets.QCheckBox('plot progress')

    def style_widgets(self):
        self.r_cutoff_txt.setMaximumWidth(80)
        self.optimize_iterations_txt.setMaximumWidth(80)

        self.attenuation_factor_sb.setRange(1, 1000)
        self.attenuation_factor_sb.setSingleStep(1)
        self.attenuation_factor_sb.setValue(1)
        self.attenuation_factor_sb.setAlignment(QtCore.Qt.AlignRight)

    def create_layout(self):
        self.main_layout = QtWidgets.QVBoxLayout()
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(5)

        self.main_layout.addWidget(self.activate_cb)
        self.main_layout.addWidget(HorizontalLine())

        self.grid_layout = QtWidgets.QGridLayout()
        self.grid_layout.setContentsMargins(0, 0, 0, 0)
        self.grid_layout.setSpacing(5)

        self.grid_layout.addWidget(self.r_cutoff_lbl, 1, 0)
        self.grid_layout.addWidget(self.r_cutoff_txt, 1, 1)
        self.grid_layout.addWidget(QtWidgets.QLabel('A'), 1, 2)

        self.grid_layout.addWidget(self.optimize_iterations_lbl, 4, 0)
        self.grid_layout.addWidget(self.optimize_iterations_txt, 4, 1)

        self.grid_layout.addWidget(self.attenuation_factor_lbl, 5, 0)
        self.grid_layout.addWidget(self.attenuation_factor_sb, 5, 1)

        self.grid_layout.addWidget(self.plot_progress_cb, 6, 1, 1, 2)

        self.param_widget = QtWidgets.QWidget()
        self.param_widget.setLayout(self.grid_layout)

        self.main_layout.addWidget(self.param_widget)

        self.setLayout(self.main_layout)

    def create_signals(self):
        self.activate_cb.stateChanged.connect(self.param_widget.setVisible)
        self.activate_cb.stateChanged.connect(
            self.emit_calculation_changed_signal)

        self.r_cutoff_txt.editingFinished.connect(
            self.emit_calculation_changed_signal)
        self.optimize_iterations_txt.editingFinished.connect(
            self.emit_calculation_changed_signal)
        self.attenuation_factor_sb.valueChanged.connect(
            self.emit_calculation_changed_signal)

    def emit_calculation_changed_signal(self):
        if self.r_cutoff_txt.isModified():
            self.r_cutoff_txt.setModified(False)
        elif self.optimize_iterations_txt.isModified():
            self.optimize_iterations_txt.setModified(False)
        self.optimization_parameters_changed.emit()

    def get_parameter(self):
        activate = self.activate_cb.isChecked()
        r_cutoff = float(str(self.r_cutoff_txt.text()))
        iterations = int(str(self.optimize_iterations_txt.text()))
        attenuation = int(self.attenuation_factor_sb.value())
        return activate, r_cutoff, iterations, attenuation

    def set_parameter(self, r_cutoff, iterations, attenuation):
        self.blockSignals(True)
        self.r_cutoff_txt.setText(str(r_cutoff))
        self.optimize_iterations_txt.setText(str(int(iterations)))
        self.attenuation_factor_sb.setValue(attenuation)
        self.blockSignals(False)



# -*- coding: utf-8 -*-
import re
from typing import Optional
from copy import copy

import numpy as np
from scipy import interpolate
import lmfit

from .scattering_factors import calculate_coherent_scattering_factor, calculate_incoherent_scattered_intensity
from . import Pattern
from . import scattering_factors

__all__ = ['calculate_f_mean_squared', 'calculate_f_squared_mean', 'calculate_incoherent_scattering',
           'extrapolate_to_zero_linear', 'extrapolate_to_zero_poly', 'extrapolate_to_zero_spline', 'calculate_s0',
           'extrapolate_to_zero_step', 'convert_density_to_atoms_per_cubic_angstrom', 'normalize_composition',
           'convert_two_theta_to_q_space', 'convert_two_theta_to_q_space_raw', 'calculate_weighting_factor']


def calculate_f_mean_squared(composition: dict, q: np.ndarray, sf_source='hajdu') -> np.ndarray:
    """
    Calculates the square of the mean form factor for a given composition over q.

    :param composition: dictionary with elements as key and abundances as relative numbers
    :param q: Q value or numpy array with a unit of A^-1
    :param sf_source: source of the scattering factors. Possible sources are 'hajdu' and 'brown_hubbell'.
    """
    norm_elemental_abundances = normalize_composition(composition)

    res = 0
    for element, amount in norm_elemental_abundances.items():
        res += amount * calculate_coherent_scattering_factor(element, q, sf_source)
    return res ** 2


def calculate_f_squared_mean(composition: dict[str, float], q: np.ndarray, sf_source: str = 'hajdu') -> np.ndarray:
    """
    Calculates the mean of the squared form factors for a given composition for a given q vector.

    :param composition: dictionary with elements as key and abundances as relative numbers
    :param q: Q value or numpy array with a unit of A^-1
    :param sf_source: source of the scattering factors. Possible sources are 'hajdu' and 'brown_hubbell'.

    :return: mean of the squared form factors
    """
    norm_elemental_abundances = normalize_composition(composition)

    res = 0
    for key, value in norm_elemental_abundances.items():
        res += value * calculate_coherent_scattering_factor(key, q, sf_source) ** 2
    return res


def calculate_incoherent_scattering(composition: dict[str, float], q: np.ndarray, sf_source: str = 'hajdu') \
        -> np.ndarray:
    """
    Calculates compton/incoherent scattering for a given composition

    :param composition: dictionary with elements as key and abundances as relative numbers
    :param q: Q value or numpy array with a unit of A^-1
    :param sf_source: source of the scattering factors. Possible sources are 'hajdu' and 'brown_hubbell'.

    :return: incoherent scattering array
    """
    norm_elemental_abundances = normalize_composition(composition)

    res = 0
    for key, value in norm_elemental_abundances.items():
        res += value * calculate_incoherent_scattered_intensity(key, q, sf_source)
    return res


def calculate_s0(composition: dict[str, float], sf_source: str = 'hajdu') -> float:
    """
    Calculates the I0 value for a given composition by extrapolating the coherent scattering factor to zero where I(Q)
    and the Compton scattering should have zero intensities.

    :param composition: dictionary with elements as keys and abundances as relative numbers
    :param sf_source: source of the scattering factors. Possible sources are 'hajdu' and 'brown_hubbell'.

    :return: I0 value
    """
    f_mean_squared = calculate_f_mean_squared(composition, np.array([0]), sf_source)
    f_squared_mean = calculate_f_squared_mean(composition, np.array([0]), sf_source)

    return float(-f_squared_mean / f_mean_squared + 1)


def calculate_weighting_factor(composition: dict[str, float], element_1: str, element_2: str, q: np.ndarray,
                               sf_source='hajdu'):
    """
    Calculates the weighting factor for an element-element contribution in a given composition (e.g. for Si-O in SiO2)

    :param composition: dictionary with elements as key string and abundances as relative numbers
    :param element_1: string giving element 1
    :param element_2: string giving element 2
    :param q: Q value or numpy array with a unit of A^-1
    :param sf_source: source of the scattering factors. Possible sources are 'hajdu' and 'brown_hubbell'.

    :return: weighting factor array
    """
    if element_1 == element_2:
        factor = 1
    else:
        factor = 2

    num_atoms = sum([val for _, val in composition.items()])
    f = {}  # form factors
    c = {}  # concentrations
    for element, val in composition.items():
        f[element] = calculate_coherent_scattering_factor(element, q, sf_source)
        c[element] = val / num_atoms

    f_sum_squared = np.zeros_like(q)
    for element, conc in c.items():
        f_sum_squared += f[element] * conc
    f_sum_squared = f_sum_squared ** 2

    return factor * c[element_1] * c[element_2] * f[element_1] * f[element_2] / f_sum_squared


def normalize_composition(composition):
    """
    normalizes elemental abundances to 1

    :param composition: dictionary with elements as key and abundances as relative numbers
    :return: normalized elemental abundances dictionary
    """
    sum = 0.0
    for key, val in composition.items():
        sum += val

    result = copy(composition)

    for key in result:
        result[key] /= sum

    return result


def convert_density_to_atoms_per_cubic_angstrom(composition, density):
    """
    Converts densities given in g/cm3 into atoms per A^3

    :param composition: dictionary with elements as key and abundances as relative numbers
    :param density: density in g/cm^3
    :return: density in atoms/A^3
    """

    # get_smallest abundance
    norm_elemental_abundances = normalize_composition(composition)
    mean_z = 0.0
    for element, concentration in norm_elemental_abundances.items():
        element = re.findall('[A-zA-Z]*', element)[0]
        mean_z += concentration * scattering_factors.atomic_weights['AW'][element]
    return density / mean_z * .602214129


def extrapolate_to_zero_step(pattern: Pattern, y0=0) -> Pattern:
    """
    Extrapolates a pattern to (0, y0) by setting everything below the q_min of the pattern to y0 (default=0)

    :param pattern: input Pattern
    :param y0: y value at x = 0

    :return: extrapolated Pattern
    """
    x, y = pattern.data
    step = x[1] - x[0]
    low_x = np.arange(min(x), 0 - step / 2, -step)[::-1]
    low_y = np.zeros(low_x.shape) + y0

    return Pattern(np.concatenate((low_x, x)),
                   np.concatenate((low_y, y)))


def extrapolate_to_zero_linear(pattern: Pattern, y0=0) -> Pattern:
    """
    Extrapolates a pattern to (0, y0) using a linear function from the leftest point in the pattern

    :param pattern: input Pattern
    :param y0: y value at x = 0

    :return: new extrapolated Pattern (includes the original data)
    """
    x, y = pattern.data
    step = x[1] - x[0]
    low_x = np.arange(min(x), 0 - step / 2, -step)[::-1]
    low_y = (y[0] - y0) / x[0] * low_x + y0
    return Pattern(np.concatenate((low_x, x)),
                   np.concatenate((low_y, y)))


def extrapolate_to_zero_spline(pattern: Pattern,
                               x_max: float,
                               smooth_factor: Optional[float] = None,
                               replace: bool = False,
                               y0: float = 0) -> Pattern:
    """
    Extrapolates a pattern to (0, y0) using a spline function.
    If the spline hits zero on the y-axis at an x value higher than 0 all values below this intersection
    will be set to zero

    :param pattern: input pattern
    :param x_max: defines the maximum x value within the spline will be fitted to the input pattern, This parameter
    should be larger than the minimum of the pattern x
    :param smooth_factor: defines the smoothing of the spline extrapolation please see numpy.UnivariateSpline manual for
    explanations
    :param replace: boolean flag whether to replace the data values in the fitted region (default = False)
    :param y0: y value at x = 0

    :return: extrapolated Pattern (includes the original one)
    """

    x, y = pattern.data
    x_step = x[1] - x[0]
    x_low = np.arange(min(x), 0 - x_step / 2, -x_step)[::-1]

    x_intersection = np.concatenate(([0], x[x < x_max]))
    y_intersection = np.concatenate(([y0], y[x < x_max]))

    if replace:
        x_low = np.concatenate((x_low, x_intersection[1:]))
        ind = x > x_max
        x = x[ind]
        y = y[ind]

    spl = interpolate.UnivariateSpline(x_intersection, y_intersection, s=smooth_factor)
    y_low = spl(x_low)

    ind_below_zero = np.where(y_low < y0)[0]

    if len(ind_below_zero) > 0:
        y_low[:ind_below_zero[-1]] = y0

    return Pattern(np.concatenate((x_low, x)),
                   np.concatenate((y_low, y)))


def extrapolate_to_zero_poly(pattern: Pattern, x_max: float, replace: bool = False, y0: float = 0) -> Pattern:
    """
    Extrapolates a pattern to (0, y0) using a 2nd order polynomial:

    .. math::
        y = a*(x-c)+b*(x-c)^2 + y0


    if the polynomial extrapolation hits the value of y0 (default=0) at an x value higher than zero all y values below
    this intersection will be set to y0.

    :param pattern: input pattern
    :param x_max: defines the maximum x value within the polynomial will be fit
    :param replace: boolean flag whether to replace the data values in the fitted region (default = False)
    :param y0: y value at x = 0

    :return: extrapolated Pattern
    """

    x, y = pattern.data
    x_step = x[1] - x[0]

    x_fit = x[x < x_max]
    y_fit = y[x < x_max]

    params = lmfit.Parameters()
    params.add("a", value=1, min=0)
    params.add("b", value=1, min=0)
    params.add("c", value=1)

    def optimization_fcn(params):
        a = params['a'].value
        b = params['b'].value
        c = params['c'].value

        return y_fit - (x_fit - c) * a - (x_fit - c) ** 2 * b + y0

    result = lmfit.minimize(optimization_fcn, params)
    a = result.params['a'].value
    b = result.params['b'].value
    c = result.params['c'].value

    x_low = np.arange(min(x), 0, -x_step)[::-1]

    if replace:
        x_low = np.concatenate((x_low, x_fit[1:]))
        ind = x > x_max
        x = x[ind]
        y = y[ind]

    y_low = a * (x_low - c) + b * (x_low - c) ** 2 - y0
    y_low[y_low < y0] = y0

    return Pattern(np.concatenate((x_low, x)),
                   np.concatenate((y_low, y)))


def convert_two_theta_to_q_space_raw(two_theta, wavelength):
    """
    Converts two theta values into q space
    """
    return 4 * np.pi * np.sin(two_theta / 360.0 * np.pi) / wavelength


def convert_two_theta_to_q_space(pattern, wavelength):
    """
    Returns a new pattern with the x-axis converted from two theta into q space
    """
    q_pattern = copy(pattern)
    q_pattern._x = convert_two_theta_to_q_space_raw(q_pattern.x, wavelength)
    return q_pattern



# -*- coding: utf-8 -*-

from qtpy import QtCore, QtWidgets

from ..custom import HorizontalLine, ValueLabelTxtPair


class SollerWidget(QtWidgets.QWidget):
    soller_parameters_changed = QtCore.Signal()

    def __init__(self, *args):
        super(SollerWidget, self).__init__(*args)

        self.create_layout_and_widgets()
        self.style_widgets()
        self.create_signals()

        self.param_widget.setVisible(False)
        self.activate_cb.setChecked(False)

    def create_layout_and_widgets(self):
        self.main_layout = QtWidgets.QVBoxLayout()

        self.activate_cb = QtWidgets.QCheckBox("activate")
        self.main_layout.addWidget(self.activate_cb)
        self.main_layout.addWidget(HorizontalLine())

        self.param_layout = QtWidgets.QGridLayout()
        self.thickness_txt = ValueLabelTxtPair("Sample thickness:", 0.2, "mm", self.param_layout, 0)
        self.wavelength_txt = ValueLabelTxtPair("X-ray wavelength:", 0.31, "A", self.param_layout, 1)
        self.param_layout.addWidget(HorizontalLine(), 2, 0, 1, 3)
        self.inner_radius_txt = ValueLabelTxtPair("Inner radius:", '', "mm", self.param_layout, 4)
        self.outer_radius_txt = ValueLabelTxtPair("Outer radius:", '', "mm", self.param_layout, 5)
        self.inner_width_txt = ValueLabelTxtPair("Inner width:", '', "mm", self.param_layout, 6)
        self.outer_width_txt = ValueLabelTxtPair("Outer width:", '', "mm", self.param_layout, 7)
        self.inner_length_txt = ValueLabelTxtPair("Inner length:", '', "mm", self.param_layout, 8)
        self.outer_length_txt = ValueLabelTxtPair("Inner length:", '', "mm", self.param_layout, 9)

        self.param_widget = QtWidgets.QWidget()
        self.param_widget.setLayout(self.param_layout)

        self.main_layout.addWidget(self.param_widget)
        self.setLayout(self.main_layout)

    def style_widgets(self):
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(5)

        self.param_layout.setContentsMargins(50, 0, 0, 0)
        self.param_layout.setVerticalSpacing(7)

    #

    def create_signals(self):
        self.activate_cb.stateChanged.connect(self.param_widget.setVisible)
        self.activate_cb.stateChanged.connect(self.emit_parameters_changed_signal)

        self.thickness_txt.editingFinished.connect(self.emit_parameters_changed_signal)
        self.wavelength_txt.editingFinished.connect(self.emit_parameters_changed_signal)
        self.inner_radius_txt.editingFinished.connect(self.emit_parameters_changed_signal)
        self.outer_radius_txt.editingFinished.connect(self.emit_parameters_changed_signal)
        self.inner_width_txt.editingFinished.connect(self.emit_parameters_changed_signal)
        self.outer_width_txt.editingFinished.connect(self.emit_parameters_changed_signal)
        self.inner_length_txt.editingFinished.connect(self.emit_parameters_changed_signal)
        self.outer_length_txt.editingFinished.connect(self.emit_parameters_changed_signal)

    def emit_parameters_changed_signal(self, *_, **__):
        self.soller_parameters_changed.emit()

    def get_parameters(self):
        return {"sample_thickness": self.thickness_txt.get_value(),
                "wavelength": self.wavelength_txt.get_value(),
                "inner_radius": self.inner_radius_txt.get_value(),
                "outer_radius": self.outer_radius_txt.get_value(),
                "inner_width": self.inner_width_txt.get_value(),
                "outer_width": self.outer_width_txt.get_value(),
                "inner_length": self.inner_length_txt.get_value(),
                "outer_length": self.outer_length_txt.get_value()}

    def set_parameters(self, parameter):
        self.blockSignals(True)
        self.thickness_txt.set_value(parameter["sample_thickness"])
        self.wavelength_txt.set_value(parameter["wavelength"])
        self.inner_radius_txt.set_value(parameter["inner_radius"])
        self.outer_radius_txt.set_value(parameter["outer_radius"])
        self.inner_width_txt.set_value(parameter["inner_width"])
        self.outer_width_txt.set_value(parameter["outer_width"])
        self.inner_length_txt.set_value(parameter["inner_length"])
        self.outer_length_txt.set_value(parameter["outer_length"])
        self.blockSignals(False)
